# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import unicodedata

from .tokenizer_utils import (
    PretrainedTokenizer,
    _is_control,
    _is_punctuation,
    _is_symbol,
    _is_whitespace,
    convert_to_unicode,
    whitespace_tokenize,
)

__all__ = [
    "BasicTokenizer",
    "BertTokenizer",
    "WordpieceTokenizer",
]


class BasicTokenizer(object):
    """
    Runs basic tokenization (punctuation splitting, lower casing, etc.).

    Args:
        do_lower_case (bool):
            Whether to lowercase the input when tokenizing.
            Defaults to `True`.
        never_split (Iterable):
            Collection of tokens which will never be split during tokenization. Only has an effect when
            `do_basic_tokenize=True`
        tokenize_chinese_chars (bool):
            Whether to tokenize Chinese characters.
        strip_accents: (bool):
            Whether to strip all accents. If this option is not specified, then it will be determined by the
            value for `lowercase` (as in the original BERT).
    """

    def __init__(
        self,
        do_lower_case=True,
        never_split=None,
        tokenize_chinese_chars=True,
        strip_accents=None,
    ):
        """Constructs a BasicTokenizer."""
        if never_split is None:
            never_split = []
        self.do_lower_case = do_lower_case
        self.never_split = set(never_split)
        self.tokenize_chinese_chars = tokenize_chinese_chars
        self.strip_accents = strip_accents

    def tokenize(self, text, never_split=None):
        """
        Tokenizes a piece of text using basic tokenizer.

        Args:
            text (str): A piece of text.
            never_split (List[str]): List of token not to split.

        Returns:
            list(str): A list of tokens.

        Examples:
            .. code-block::

                from paddlenlp.transformers import BasicTokenizer
                basictokenizer = BasicTokenizer()
                tokens = basictokenizer.tokenize('He was a puppeteer')
                '''
                ['he', 'was', 'a', 'puppeteer']
                '''
        """
        text = convert_to_unicode(text)
        never_split = (
            self.never_split.union(set(never_split))
            if never_split
            else self.never_split
        )
        text = self._clean_text(text)

        if self.tokenize_chinese_chars:
            text = self._tokenize_chinese_chars(text)
        orig_tokens = whitespace_tokenize(text)
        split_tokens = []
        for token in orig_tokens:
            if token not in never_split:
                if self.do_lower_case:
                    token = token.lower()
                    if self.strip_accents is not False:
                        token = self._run_strip_accents(token)
                elif self.strip_accents:
                    token = self._run_strip_accents(token)
            split_tokens.extend(self._run_split_on_punc(token, never_split))

        output_tokens = whitespace_tokenize(" ".join(split_tokens))
        return output_tokens

    def _run_strip_accents(self, text):
        """
        Strips accents from a piece of text.
        """
        text = unicodedata.normalize("NFD", text)
        output = []
        for char in text:
            cat = unicodedata.category(char)
            if cat == "Mn":
                continue
            output.append(char)
        return "".join(output)

    def _run_split_on_punc(self, text, never_split=None):
        """
        Splits punctuation on a piece of text.
        """
        if never_split is not None and text in never_split:
            return [text]
        chars = list(text)
        i = 0
        start_new_word = True
        output = []
        while i < len(chars):
            char = chars[i]
            # punctuation and symbol should be treat as single char.
            if _is_punctuation(char) or _is_symbol(char):
                output.append([char])
                start_new_word = True
            else:
                if start_new_word:
                    output.append([])
                start_new_word = False
                output[-1].append(char)
            i += 1

        return ["".join(x) for x in output]

    def _tokenize_chinese_chars(self, text):
        """
        Adds whitespace around any CJK character.
        """
        output = []
        for char in text:
            cp = ord(char)
            if self._is_chinese_char(cp):
                output.append(" ")
                output.append(char)
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)

    def _is_chinese_char(self, cp):
        """
        Checks whether CP is the codepoint of a CJK character.
        """
        # This defines a "chinese character" as anything in the CJK Unicode block:
        #     https://en.wikipedia.org/wiki/CJK_Unified_Ideographs_(Unicode_block)
        #
        # Note that the CJK Unicode block is NOT all Japanese and Korean characters,
        # despite its name. The modern Korean Hangul alphabet is a different block,
        # as is Japanese Hiragana and Katakana. Those alphabets are used to write
        # space-separated words, so they are not treated specially and handled
        # like the all the other languages.
        if (
            (cp >= 0x4E00 and cp <= 0x9FFF)
            or (cp >= 0x3400 and cp <= 0x4DBF)  #
            or (cp >= 0x20000 and cp <= 0x2A6DF)  #
            or (cp >= 0x2A700 and cp <= 0x2B73F)  #
            or (cp >= 0x2B740 and cp <= 0x2B81F)  #
            or (cp >= 0x2B820 and cp <= 0x2CEAF)  #
            or (cp >= 0xF900 and cp <= 0xFAFF)
            or (cp >= 0x2F800 and cp <= 0x2FA1F)  #
        ):  #
            return True

        return False

    def _clean_text(self, text):
        """
        Performs invalid character removal and whitespace cleanup on text.
        """
        output = []
        for char in text:
            cp = ord(char)
            if cp == 0 or cp == 0xFFFD or _is_control(char):
                continue
            if _is_whitespace(char):
                output.append(" ")
            else:
                output.append(char)
        return "".join(output)


class WordpieceTokenizer(object):
    """
    Runs WordPiece tokenization.

    Args:
        vocab (Vocab|dict):
            Vocab of the word piece tokenizer.
        unk_token (str):
            A specific token to replace all unknown tokens.
        max_input_chars_per_word (int):
            If a word's length is more than
            max_input_chars_per_word, it will be dealt as unknown word.
            Defaults to 100.
    """

    def __init__(self, vocab, unk_token, max_input_chars_per_word=100):
        self.vocab = vocab
        self.unk_token = unk_token
        self.max_input_chars_per_word = max_input_chars_per_word

    def tokenize(self, text):
        """
        Tokenizes a piece of text into its word pieces.
        This uses a greedy longest-match-first algorithm to perform tokenization
        using the given vocabulary.

        Args:
            text: A single token or whitespace separated tokens. This should have
                already been passed through `BasicTokenizer`.

        Returns:
            list (str): A list of wordpiece tokens.

        Examples:
            .. code-block::

                from paddlenlp.transformers import BertTokenizer, WordpieceTokenizer

                berttokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
                vocab  = berttokenizer.vocab
                unk_token = berttokenizer.unk_token

                wordpiecetokenizer = WordpieceTokenizer(vocab,unk_token)
                inputs = wordpiecetokenizer.tokenize("unaffable")
                print(inputs)
                '''
                ["un", "##aff", "##able"]
                '''
        """

        output_tokens = []
        for token in whitespace_tokenize(text):
            chars = list(token)
            if len(chars) > self.max_input_chars_per_word:
                output_tokens.append(self.unk_token)
                continue

            is_bad = False
            start = 0
            sub_tokens = []
            while start < len(chars):
                end = len(chars)
                cur_substr = None
                while start < end:
                    substr = "".join(chars[start:end])
                    if start > 0:
                        substr = "##" + substr
                    if substr in self.vocab:
                        cur_substr = substr
                        break
                    end -= 1
                if cur_substr is None:
                    is_bad = True
                    break
                sub_tokens.append(cur_substr)
                start = end

            if is_bad:
                output_tokens.append(self.unk_token)
            else:
                output_tokens.extend(sub_tokens)
        return output_tokens


class BertTokenizer(PretrainedTokenizer):
    """
    Constructs a BERT tokenizer. It uses a basic tokenizer to do punctuation
    splitting, lower casing and so on, and follows a WordPiece tokenizer to
    tokenize as subwords.

    Args:
        vocab_file (str):
            The vocabulary file path (ends with '.txt') required to instantiate
            a `WordpieceTokenizer`.
        do_lower_case (bool, optional):
            Whether to lowercase the input when tokenizing.
            Defaults to `True`.
        do_basic_tokenize (bool, optional):
            Whether to use a basic tokenizer before a WordPiece tokenizer.
            Defaults to `True`.
        never_split (Iterable, optional):
            Collection of tokens which will never be split during tokenization. Only has an effect when
            `do_basic_tokenize=True`. Defaults to `None`.
        unk_token (str, optional):
            A special token representing the *unknown (out-of-vocabulary)* token.
            An unknown token is set to be `unk_token` inorder to be converted to an ID.
            Defaults to "[UNK]".
        sep_token (str, optional):
            A special token separating two different sentences in the same input.
            Defaults to "[SEP]".
        pad_token (str, optional):
            A special token used to make arrays of tokens the same size for batching purposes.
            Defaults to "[PAD]".
        cls_token (str, optional):
            A special token used for sequence classification. It is the last token
            of the sequence when built with special tokens. Defaults to "[CLS]".
        mask_token (str, optional):
            A special token representing a masked token. This is the token used
            in the masked language modeling task which the model tries to predict the original unmasked ones.
            Defaults to "[MASK]".
        tokenize_chinese_chars (bool, optional):
            Whether to tokenize Chinese characters.
            Defaults to `True`.
        strip_accents: (bool, optional):
            Whether to strip all accents. If this option is not specified, then it will be determined by the
            value for `lowercase` (as in the original BERT).
            Defaults to `None`.

    Examples:
        .. code-block::

            from paddlenlp.transformers import BertTokenizer
            tokenizer = BertTokenizer.from_pretrained('bert-base-uncased')

            inputs = tokenizer('He was a puppeteer')
            print(inputs)

            '''
            {'input_ids': [101, 2002, 2001, 1037, 13997, 11510, 102], 'token_type_ids': [0, 0, 0, 0, 0, 0, 0]}
            '''
    """

    resource_files_names = {"vocab_file": "vocab.txt"}  # for save_pretrained
    pretrained_resource_files_map = {
        "vocab_file": {
            "bert-base-uncased": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-uncased-vocab.txt",
            "bert-large-uncased": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-large-uncased-vocab.txt",
            "bert-base-cased": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-cased-vocab.txt",
            "bert-large-cased": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-large-cased-vocab.txt",
            "bert-base-multilingual-uncased": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-multilingual-uncased-vocab.txt",
            "bert-base-multilingual-cased": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-multilingual-cased-vocab.txt",
            "bert-base-chinese": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-chinese-vocab.txt",
            "bert-wwm-chinese": "http://bj.bcebos.com/paddlenlp/models/transformers/bert/bert-wwm-chinese-vocab.txt",
            "bert-wwm-ext-chinese": "http://bj.bcebos.com/paddlenlp/models/transformers/bert/bert-wwm-ext-chinese-vocab.txt",
            "macbert-large-chinese": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-chinese-vocab.txt",
            "macbert-base-chinese": "https://bj.bcebos.com/paddle-hapi/models/bert/bert-base-chinese-vocab.txt",
            "simbert-base-chinese": "https://bj.bcebos.com/paddlenlp/models/transformers/simbert/vocab.txt",
            "uer/chinese-roberta-base": "https://bj.bcebos.com/paddlenlp/models/transformers/uer/chinese_roberta_vocab.txt",
            "uer/chinese-roberta-medium": "https://bj.bcebos.com/paddlenlp/models/transformers/uer/chinese_roberta_vocab.txt",
            "uer/chinese-roberta-6l-768h": "https://bj.bcebos.com/paddlenlp/models/transformers/uer/chinese_roberta_vocab.txt",
            "uer/chinese-roberta-small": "https://bj.bcebos.com/paddlenlp/models/transformers/uer/chinese_roberta_vocab.txt",
            "uer/chinese-roberta-mini": "https://bj.bcebos.com/paddlenlp/models/transformers/uer/chinese_roberta_vocab.txt",
            "uer/chinese-roberta-tiny": "https://bj.bcebos.com/paddlenlp/models/transformers/uer/chinese_roberta_vocab.txt",
        }
    }
    pretrained_init_configuration = {
        "bert-base-uncased": {"do_lower_case": True},
        "bert-large-uncased": {"do_lower_case": True},
        "bert-base-cased": {"do_lower_case": False},
        "bert-large-cased": {"do_lower_case": False},
        "bert-base-multilingual-uncased": {"do_lower_case": True},
        "bert-base-multilingual-cased": {"do_lower_case": False},
        "bert-base-chinese": {"do_lower_case": False},
        "bert-wwm-chinese": {"do_lower_case": False},
        "bert-wwm-ext-chinese": {"do_lower_case": False},
        "macbert-large-chinese": {"do_lower_case": False},
        "macbert-base-chinese": {"do_lower_case": False},
        "simbert-base-chinese": {"do_lower_case": True},
        "uer/chinese-roberta-base": {"do_lower_case": True},
        "uer/chinese-roberta-medium": {"do_lower_case": True},
        "uer/chinese-roberta-6l-768h": {"do_lower_case": True},
        "uer/chinese-roberta-small": {"do_lower_case": True},
        "uer/chinese-roberta-mini": {"do_lower_case": True},
        "uer/chinese-roberta-tiny": {"do_lower_case": True},
    }
    max_model_input_sizes = {
        "bert-base-uncased": 512,
        "bert-large-uncased": 512,
        "bert-base-cased": 512,
        "bert-large-cased": 512,
        "bert-base-multilingual-uncased": 512,
        "bert-base-multilingual-cased": 512,
        "bert-base-chinese": 512,
        "bert-wwm-chinese": 512,
        "bert-wwm-ext-chinese": 512,
        "macbert-large-chinese": 512,
        "macbert-base-chinese": 512,
        "simbert-base-chinese": 512,
        "uer/chinese-roberta-base": 512,
        "uer/chinese-roberta-medium": 512,
        "uer/chinese-roberta-6l-768h": 512,
        "uer/chinese-roberta-small": 512,
        "uer/chinese-roberta-mini": 512,
        "uer/chinese-roberta-tiny": 512,
    }
    padding_side = "right"

    def __init__(
        self,
        vocab_file,
        do_lower_case=True,
        do_basic_tokenize=True,
        never_split=None,
        unk_token="[UNK]",
        sep_token="[SEP]",
        pad_token="[PAD]",
        cls_token="[CLS]",
        mask_token="[MASK]",
        tokenize_chinese_chars=True,
        strip_accents=None,
        **kwargs
    ):

        if not os.path.isfile(vocab_file):
            raise ValueError(
                "Can't find a vocabulary file at path '{}'. To load the "
                "vocabulary from a pretrained model please use "
                "`tokenizer = BertTokenizer.from_pretrained(PRETRAINED_MODEL_NAME)`".format(
                    vocab_file
                )
            )
        self.do_lower_case = do_lower_case
        self.vocab = self.load_vocabulary(vocab_file, unk_token=unk_token)
        self.do_basic_tokenize = do_basic_tokenize
        if do_basic_tokenize:
            self.basic_tokenizer = BasicTokenizer(
                do_lower_case=do_lower_case,
                never_split=never_split,
                tokenize_chinese_chars=tokenize_chinese_chars,
                strip_accents=strip_accents,
            )
        self.wordpiece_tokenizer = WordpieceTokenizer(
            vocab=self.vocab, unk_token=unk_token
        )

    @property
    def vocab_size(self):
        """
        Return the size of vocabulary.

        Returns:
            int: The size of vocabulary.
        """

        return len(self.vocab)

    def get_vocab(self):
        return dict(self.vocab.token_to_idx, **self.added_tokens_encoder)

    def _tokenize(self, text):
        """
        End-to-end tokenization for BERT models.

        Args:
            text (str): The text to be tokenized.

        Returns:
            list: A list of string representing converted tokens.
        """
        split_tokens = []
        if self.do_basic_tokenize:
            for token in self.basic_tokenizer.tokenize(
                text, never_split=self.all_special_tokens
            ):
                # If the token is part of the never_split set
                if token in self.basic_tokenizer.never_split:
                    split_tokens.append(token)
                else:
                    split_tokens += self.wordpiece_tokenizer.tokenize(token)
        else:
            split_tokens = self.wordpiece_tokenizer.tokenize(text)
        return split_tokens

    def convert_tokens_to_string(self, tokens):
        """
        Converts a sequence of tokens (list of string) to a single string. Since
        the usage of WordPiece introducing `##` to concat subwords, also removes
        `##` when converting.

        Args:
            tokens (list): A list of string representing tokens to be converted.

        Returns:
            str: Converted string from tokens.

        Examples:
            .. code-block::

                from paddlenlp.transformers import BertTokenizer

                berttokenizer = BertTokenizer.from_pretrained('bert-base-uncased')
                tokens = berttokenizer.tokenize('He was a puppeteer')
                '''
                ['he', 'was', 'a', 'puppet', '##eer']
                '''
                strings = tokenizer.convert_tokens_to_string(tokens)
                '''
                he was a puppeteer
                '''
        """

        out_string = " ".join(tokens).replace(" ##", "").strip()
        return out_string

    def num_special_tokens_to_add(self, pair=False):
        """
        Returns the number of added tokens when encoding a sequence with special tokens.

        Args:
            pair(bool):
                Whether the input is a sequence pair or a single sequence.
                Defaults to `False` and the input is a single sequence.

        Returns:
            int: Number of tokens added to sequences.
        """
        token_ids_0 = []
        token_ids_1 = []
        return len(
            self.build_inputs_with_special_tokens(
                token_ids_0, token_ids_1 if pair else None
            )
        )

    def build_inputs_with_special_tokens(self, token_ids_0, token_ids_1=None):
        """
        Build model inputs from a sequence or a pair of sequence for sequence classification tasks by concatenating and
        adding special tokens.

        A BERT sequence has the following format:

        - single sequence:      ``[CLS] X [SEP]``
        - pair of sequences:        ``[CLS] A [SEP] B [SEP]``

        Args:
            token_ids_0 (List[int]):
                List of IDs to which the special tokens will be added.
            token_ids_1 (List[int], optional):
                Optional second list of IDs for sequence pairs. Defaults to None.

        Returns:
            List[int]: List of input_id with the appropriate special tokens.
        """
        if token_ids_1 is None:
            return [self.cls_token_id] + token_ids_0 + [self.sep_token_id]
        _cls = [self.cls_token_id]
        _sep = [self.sep_token_id]
        return _cls + token_ids_0 + _sep + token_ids_1 + _sep

    def build_offset_mapping_with_special_tokens(
        self, offset_mapping_0, offset_mapping_1=None
    ):
        """
        Build offset map from a pair of offset map by concatenating and adding offsets of special tokens.

        A BERT offset_mapping has the following format:

        - single sequence:      ``(0,0) X (0,0)``
        - pair of sequences:        ``(0,0) A (0,0) B (0,0)``

        Args:
            offset_mapping_ids_0 (List[tuple]):
                List of wordpiece offsets to which the special tokens will be added.
            offset_mapping_ids_1 (List[tuple], optional):
                Optional second list of wordpiece offsets for offset mapping pairs. Defaults to None.

        Returns:
            List[tuple]: A list of wordpiece offsets with the appropriate offsets of special tokens.
        """
        if offset_mapping_1 is None:
            return [(0, 0)] + offset_mapping_0 + [(0, 0)]

        return [(0, 0)] + offset_mapping_0 + [(0, 0)] + offset_mapping_1 + [(0, 0)]

    def create_token_type_ids_from_sequences(self, token_ids_0, token_ids_1=None):
        """
        Create a mask from the two sequences passed to be used in a sequence-pair classification task.

        A BERT sequence pair mask has the following format:
        ::

            0 0 0 0 0 0 0 0 0 0 0 1 1 1 1 1 1 1 1 1
            | first sequence    | second sequence |

        If `token_ids_1` is `None`, this method only returns the first portion of the mask (0s).

        Args:
            token_ids_0 (List[int]):
                A list of `inputs_ids` for the first sequence.
            token_ids_1 (List[int], optional):
                Optional second list of IDs for sequence pairs. Defaults to None.

        Returns:
            List[int]: List of token_type_id according to the given sequence(s).
        """
        _sep = [self.sep_token_id]
        _cls = [self.cls_token_id]
        if token_ids_1 is None:
            return len(_cls + token_ids_0 + _sep) * [0]
        return len(_cls + token_ids_0 + _sep) * [0] + len(token_ids_1 + _sep) * [1]

    def get_special_tokens_mask(
        self, token_ids_0, token_ids_1=None, already_has_special_tokens=False
    ):
        """
        Retrieves sequence ids from a token list that has no special tokens added. This method is called when adding
        special tokens using the tokenizer ``encode`` methods.

        Args:
            token_ids_0 (List[int]):
                A list of `inputs_ids` for the first sequence.
            token_ids_1 (List[int], optional):
                Optional second list of IDs for sequence pairs. Defaults to None.
            already_has_special_tokens (bool, optional): Whether or not the token list is already
                formatted with special tokens for the model. Defaults to None.

        Returns:
            List[int]: The list of integers either be 0 or 1: 1 for a special token, 0 for a sequence token.
        """

        if already_has_special_tokens:
            if token_ids_1 is not None:
                raise ValueError(
                    "You should not supply a second sequence if the provided sequence of "
                    "ids is already formatted with special tokens for the model."
                )
            return list(
                map(lambda x: 1 if x in self.all_special_ids else 0, token_ids_0)
            )

        if token_ids_1 is not None:
            return [1] + ([0] * len(token_ids_0)) + [1] + ([0] * len(token_ids_1)) + [1]
        return [1] + ([0] * len(token_ids_0)) + [1]

    def _convert_id_to_token(self, index):
        """Converts an index (integer) in a token (str) using the vocab."""
        return self.vocab._idx_to_token.get(index, self.unk_token)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json
import os
import os.path as osp
from collections import Counter, defaultdict
from pathlib import Path

from PIL import Image

from ...base import BaseDatasetChecker
from ..model_list import MODELS
from .dataset_src import check, convert, deep_analyse, split_dataset


class TextRecDatasetChecker(BaseDatasetChecker):
    """Dataset Checker for Text Recognition Model"""

    entities = MODELS
    sample_num = 10

    def get_dataset_root(self, dataset_dir: str) -> str:
        """find the dataset root dir

        Args:
            dataset_dir (str): the directory that contain dataset.

        Returns:
            str: the root directory of dataset.
        """
        anno_dirs = list(Path(dataset_dir).glob("**/train.txt"))
        assert len(anno_dirs) == 1
        dataset_dir = anno_dirs[0].parent.as_posix()
        return dataset_dir

    def convert_dataset(self, src_dataset_dir: str) -> str:
        """convert the dataset from other type to specified type

        Args:
            src_dataset_dir (str): the root directory of dataset.

        Returns:
            str: the root directory of converted dataset.
        """
        return convert(
            self.check_dataset_config.convert.src_dataset_type, src_dataset_dir
        )

    def split_dataset(self, src_dataset_dir: str) -> str:
        """repartition the train and validation dataset

        Args:
            src_dataset_dir (str): the root directory of dataset.

        Returns:
            str: the root directory of splited dataset.
        """
        return split_dataset(
            src_dataset_dir,
            self.check_dataset_config.split.train_percent,
            self.check_dataset_config.split.val_percent,
        )

    def check_dataset(self, dataset_dir: str, sample_num: int = sample_num) -> dict:
        """check if the dataset meets the specifications and get dataset summary

        Args:
            dataset_dir (str): the root directory of dataset.
            sample_num (int): the number to be sampled.
        Returns:
            dict: dataset summary.
        """
        return check(
            dataset_dir,
            self.output,
            sample_num=10,
            dataset_type=self.get_dataset_type(),
        )

    def analyse(self, dataset_dir: str) -> dict:
        """deep analyse dataset

        Args:
            dataset_dir (str): the root directory of dataset.

        Returns:
            dict: the deep analysis results.
        """
        if self.global_config["model"] in ["LaTeX_OCR_rec"]:
            datatype = "LaTeXOCRDataset"
        else:
            datatype = "MSTextRecDataset"
        return deep_analyse(dataset_dir, self.output, datatype=datatype)

    def get_show_type(self) -> str:
        """get the show type of dataset

        Returns:
            str: show type
        """
        return "image"

    def get_dataset_type(self) -> str:
        """return the dataset type

        Returns:
            str: dataset type
        """
        if self.global_config["model"] in ["LaTeX_OCR_rec"]:
            return "LaTeXOCRDataset"
        else:
            return "MSTextRecDataset"



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

MODELS = [
    "CLIP_vit_base_patch16_224",
    "CLIP_vit_large_patch14_224",
    "ConvNeXt_tiny",
    "ConvNeXt_small",
    "ConvNeXt_base_224",
    "ConvNeXt_base_384",
    "ConvNeXt_large_224",
    "ConvNeXt_large_384",
    "MobileNetV1_x0_25",
    "MobileNetV1_x0_5",
    "MobileNetV1_x0_75",
    "MobileNetV1_x1_0",
    "MobileNetV2_x0_25",
    "MobileNetV2_x0_5",
    "MobileNetV2_x1_0",
    "MobileNetV2_x1_5",
    "MobileNetV2_x2_0",
    "MobileNetV3_large_x0_35",
    "MobileNetV3_large_x0_5",
    "MobileNetV3_large_x0_75",
    "MobileNetV3_large_x1_0",
    "MobileNetV3_large_x1_25",
    "MobileNetV3_small_x0_35",
    "MobileNetV3_small_x0_5",
    "MobileNetV3_small_x0_75",
    "MobileNetV3_small_x1_0",
    "MobileNetV3_small_x1_25",
    "MobileNetV4_conv_small",
    "MobileNetV4_conv_medium",
    "MobileNetV4_conv_large",
    "MobileNetV4_hybrid_medium",
    "MobileNetV4_hybrid_large",
    "PP-HGNet_tiny",
    "PP-HGNet_small",
    "PP-HGNet_base",
    "PP-HGNetV2-B0",
    "PP-HGNetV2-B1",
    "PP-HGNetV2-B2",
    "PP-HGNetV2-B3",
    "PP-HGNetV2-B4",
    "PP-HGNetV2-B5",
    "PP-HGNetV2-B6",
    "PP-LCNet_x0_25",
    "PP-LCNet_x0_25_textline_ori",
    "PP-LCNet_x0_35",
    "PP-LCNet_x0_5",
    "PP-LCNet_x0_75",
    "PP-LCNet_x1_0",
    "PP-LCNet_x1_0_doc_ori",
    "PP-LCNet_x1_0_textline_ori",
    "PP-LCNet_x1_5",
    "PP-LCNet_x2_0",
    "PP-LCNet_x2_5",
    "PP-LCNetV2_small",
    "PP-LCNetV2_base",
    "PP-LCNetV2_large",
    "ResNet101",
    "ResNet152",
    "ResNet18",
    "ResNet34",
    "ResNet50",
    "ResNet200_vd",
    "ResNet101_vd",
    "ResNet152_vd",
    "ResNet18_vd",
    "ResNet34_vd",
    "ResNet50_vd",
    "SwinTransformer_tiny_patch4_window7_224",
    "SwinTransformer_small_patch4_window7_224",
    "SwinTransformer_base_patch4_window7_224",
    "SwinTransformer_base_patch4_window12_384",
    "SwinTransformer_large_patch4_window7_224",
    "SwinTransformer_large_patch4_window12_384",
    "StarNet-S1",
    "StarNet-S2",
    "StarNet-S3",
    "StarNet-S4",
    "FasterNet-L",
    "FasterNet-M",
    "FasterNet-S",
    "FasterNet-T0",
    "FasterNet-T1",
    "FasterNet-T2",
    "PP-LCNet_x1_0_table_cls",
]



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List

import numpy as np
import pandas as pd

from ....utils.deps import class_requires_deps, is_dep_available
from ...utils.benchmark import benchmark

if is_dep_available("joblib"):
    import joblib


@benchmark.timeit
@class_requires_deps("joblib")
class TSDeNormalize:
    """A class to de-normalize time series prediction data using a pre-fitted scaler."""

    def __init__(self, scale_path: str, params_info: dict):
        """
        Initializes the TSDeNormalize class with a scaler and parameters information.

        Args:
            scale_path (str): The file path to the serialized scaler object.
            params_info (dict): Additional parameters information.
        """
        super().__init__()
        self.scaler = joblib.load(scale_path)
        self.params_info = params_info

    def __call__(self, preds_list: List[pd.DataFrame]) -> List[pd.DataFrame]:
        """
        Applies de-normalization to a list of prediction DataFrames.

        Args:
            preds_list (List[pd.DataFrame]): A list of DataFrames containing normalized prediction data.

        Returns:
            List[pd.DataFrame]: A list of DataFrames with de-normalized prediction data.
        """
        return [self.tsdenorm(pred) for pred in preds_list]

    def tsdenorm(self, pred: pd.DataFrame) -> pd.DataFrame:
        """
        De-normalizes a single prediction DataFrame.

        Args:
            pred (pd.DataFrame): A DataFrame containing normalized prediction data.

        Returns:
            pd.DataFrame: A DataFrame with de-normalized prediction data.
        """
        scale_cols = pred.columns.values.tolist()
        pred[scale_cols] = self.scaler.inverse_transform(pred[scale_cols])
        return pred


@benchmark.timeit
class ArraytoTS:
    """A class to convert arrays of predictions into time series format."""

    def __init__(self, info_params: Dict[str, Any]):
        """
        Initializes the ArraytoTS class with the given parameters.

        Args:
            info_params (Dict[str, Any]): Configuration parameters including target columns, frequency, and time column name.
        """
        super().__init__()
        self.info_params = info_params

    def __call__(
        self, ori_ts_list: List[Dict[str, Any]], pred_list: List[np.ndarray]
    ) -> List[pd.DataFrame]:
        """
        Converts a list of arrays to a list of time series DataFrames.

        Args:
            ori_ts_list (List[Dict[str, Any]]): Original time series data for each prediction, including past and covariate information.
            pred_list (List[np.ndarray]): List of prediction arrays corresponding to each time series in ori_ts_list.

        Returns:
            List[pd.DataFrame]: A list of DataFrames, each representing the forecasted time series.
        """
        return [
            self.arraytots(ori_ts, pred) for ori_ts, pred in zip(ori_ts_list, pred_list)
        ]

    def arraytots(self, ori_ts: Dict[str, Any], pred: np.ndarray) -> pd.DataFrame:
        """
        Converts a single array prediction to a time series DataFrame.

        Args:
            ori_ts (Dict[str, Any]): Original time series data for a single time series.
            pred (np.ndarray): Prediction array for the given time series.

        Returns:
            pd.DataFrame: A DataFrame representing the forecasted time series.

        Raises:
            ValueError: If none of the expected keys are found in ori_ts.
        """
        pred = pred[0]
        if ori_ts.get("past_target", None) is not None:
            ts = ori_ts["past_target"]
        elif ori_ts.get("observed_cov_numeric", None) is not None:
            ts = ori_ts["observed_cov_numeric"]
        elif ori_ts.get("known_cov_numeric", None) is not None:
            ts = ori_ts["known_cov_numeric"]
        elif ori_ts.get("static_cov_numeric", None) is not None:
            ts = ori_ts["static_cov_numeric"]
        else:
            raise ValueError("No value in ori_ts")

        column_name = (
            self.info_params["target_cols"]
            if "target_cols" in self.info_params
            else self.info_params["feature_cols"]
        )
        if isinstance(self.info_params["freq"], str):
            past_target_index = ts.index
            if past_target_index.freq is None:
                past_target_index.freq = pd.infer_freq(ts.index)
            future_target_index = pd.date_range(
                past_target_index[-1] + past_target_index.freq,
                periods=pred.shape[0],
                freq=self.info_params["freq"],
                name=self.info_params["time_col"],
            )
        elif isinstance(self.info_params["freq"], int):
            start_idx = max(ts.index) + 1
            stop_idx = start_idx + pred.shape[0]
            future_target_index = pd.RangeIndex(
                start=start_idx,
                stop=stop_idx,
                step=self.info_params["freq"],
                name=self.info_params["time_col"],
            )

        future_target = pd.DataFrame(
            np.reshape(pred, newshape=[pred.shape[0], -1]),
            index=future_target_index,
            columns=column_name,
        )
        return future_target



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, Optional, Union

from ....utils import logging
from ...utils.hpi import HPIConfig
from ...utils.pp_option import PaddlePredictorOption
from ..base import BasePipeline


class PP_ChatOCR_Pipeline(BasePipeline):
    """PP-ChatOCR Pipeline"""

    def __init__(
        self,
        device: str = None,
        pp_option: PaddlePredictorOption = None,
        use_hpip: bool = False,
        hpi_config: Optional[Union[Dict[str, Any], HPIConfig]] = None,
    ) -> None:
        """Initializes the pp-chatocrv3-doc pipeline.

        Args:
            config (Dict): Configuration dictionary containing various settings.
            device (str, optional): Device to run the predictions on. Defaults to None.
            pp_option (PaddlePredictorOption, optional): PaddlePredictor options. Defaults to None.
            use_hpip (bool, optional): Whether to use the high-performance
                inference plugin (HPIP) by default. Defaults to False.
            hpi_config (Optional[Union[Dict[str, Any], HPIConfig]], optional):
                The default high-performance inference configuration dictionary.
                Defaults to None.
        """

        super().__init__(
            device=device, pp_option=pp_option, use_hpip=use_hpip, hpi_config=hpi_config
        )

    def visual_predict(self):
        """
        This function takes an input image or a list of images and performs various visual
        prediction tasks such as document orientation classification, document unwarping,
        general OCR, seal recognition, and table recognition based on the provided flags.
        """

        raise NotImplementedError(
            "The method `visual_predict` has not been implemented yet."
        )

    def save_visual_info_list(self):
        """
        Save the visual info list to the specified file path.
        """
        raise NotImplementedError(
            "The method `save_visual_info_list` has not been implemented yet."
        )

    def load_visual_info_list(self):
        """
        Loads visual info list from a file.
        """
        raise NotImplementedError(
            "The method `load_visual_info_list` has not been implemented yet."
        )

    def build_vector(self):
        """
        Build a vector representation from visual information.
        """
        raise NotImplementedError(
            "The method `build_vector` has not been implemented yet."
        )

    def save_vector(self):
        """
        Save the vector information to a specified path.
        """
        raise NotImplementedError(
            "The method `save_vector` has not been implemented yet."
        )

    def load_vector(self):
        """
        Loads vector information from a file.
        """
        raise NotImplementedError(
            "The method `load_vector` has not been implemented yet."
        )

    def chat(self):
        """
        Generates chat results based on the provided key list and visual information.
        """
        raise NotImplementedError("The method `chat` has not been implemented yet.")

    def predict(self, *args, **kwargs) -> None:
        logging.error(
            "PP-ChatOCR Pipeline do not support to call `predict()` directly! Please invoke `visual_predict`, `build_vector`, `chat` sequentially to obtain the result."
        )
        return



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import List, Optional, Tuple, Type

import paddle
import paddle.nn as nn
import paddle.nn.functional as F

from ...common.vlm.transformers.model_outputs import CausalLMOutputWithPast
from .qwen2 import Qwen2Config, Qwen2ForCausalLM, Qwen2Model


class MLPBlock(paddle.nn.Layer):
    def __init__(
        self,
        embedding_dim: int,
        mlp_dim: int,
        act: Type[paddle.nn.Layer] = paddle.nn.GELU,
    ) -> None:
        super().__init__()
        self.lin1 = nn.Linear(embedding_dim, mlp_dim)
        self.lin2 = nn.Linear(mlp_dim, embedding_dim)
        self.act = act()

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        return self.lin2(self.act(self.lin1(x)))


class LayerNorm2d(paddle.nn.Layer):
    def __init__(self, num_channels: int, epsilon: float = 1e-06) -> None:
        super().__init__()
        self.weight = paddle.base.framework.EagerParamBase.from_tensor(
            tensor=paddle.ones(shape=num_channels)
        )
        self.bias = paddle.base.framework.EagerParamBase.from_tensor(
            tensor=paddle.zeros(shape=num_channels)
        )
        self.epsilon = epsilon

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        u = x.mean(axis=1, keepdim=True)
        s = (x - u).pow(y=2).mean(axis=1, keepdim=True)
        x = (x - u) / paddle.sqrt(x=s + self.epsilon)
        x = self.weight[:, None, None] * x + self.bias[:, None, None]
        return x


class ImageEncoderViT(paddle.nn.Layer):
    def __init__(
        self,
        img_size: int = 1024,
        patch_size: int = 16,
        in_chans: int = 3,
        embed_dim: int = 768,
        depth: int = 12,
        num_heads: int = 12,
        mlp_ratio: float = 4.0,
        out_chans: int = 256,
        qkv_bias: bool = True,
        norm_layer: Type[nn.Layer] = nn.LayerNorm,
        act_layer: Type[nn.Layer] = nn.GELU,
        use_abs_pos: bool = True,
        use_rel_pos: bool = False,
        rel_pos_zero_init: bool = True,
        window_size: int = 0,
        global_attn_indexes: Tuple[int, ...] = (),
    ) -> None:
        """
        Args:
            img_size (int): Input image size.
            patch_size (int): Patch size.
            in_chans (int): Number of input image channels.
            embed_dim (int): Patch embedding dimension.
            depth (int): Depth of ViT.
            num_heads (int): Number of attention heads in each ViT block.
            mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            qkv_bias (bool): If True, add a learnable bias to query, key, value.
            norm_layer (nn.Layer): Normalization layer.
            act_layer (nn.Layer): Activation layer.
            use_abs_pos (bool): If True, use absolute positional embeddings.
            use_rel_pos (bool): If True, add relative positional embeddings to the attention map.
            rel_pos_zero_init (bool): If True, zero initialize relative positional parameters.
            window_size (int): Window size for window attention blocks.
            global_attn_indexes (list): Indexes for blocks using global attention.
        """
        super().__init__()
        self.img_size = img_size

        self.patch_embed = PatchEmbed(
            kernel_size=(patch_size, patch_size),
            stride=(patch_size, patch_size),
            in_chans=in_chans,
            embed_dim=embed_dim,
        )

        self.pos_embed: Optional[paddle.base.framework.EagerParamBase.from_tensor] = (
            None
        )
        if use_abs_pos:
            self.pos_embed = paddle.base.framework.EagerParamBase.from_tensor(
                tensor=paddle.zeros(
                    shape=[1, img_size // patch_size, img_size // patch_size, embed_dim]
                )
            )

        self.blocks = paddle.nn.LayerList()
        for i in range(depth):
            block = Block(
                dim=embed_dim,
                num_heads=num_heads,
                mlp_ratio=mlp_ratio,
                qkv_bias=qkv_bias,
                norm_layer=norm_layer,
                act_layer=act_layer,
                use_rel_pos=use_rel_pos,
                rel_pos_zero_init=rel_pos_zero_init,
                window_size=window_size if i not in global_attn_indexes else 0,
                input_size=(img_size // patch_size, img_size // patch_size),
            )
            self.blocks.append(block)

        self.neck = nn.Sequential(
            nn.Conv2D(
                embed_dim,
                out_chans,
                kernel_size=1,
                bias_attr=False,
            ),
            LayerNorm2d(out_chans),
            nn.Conv2D(
                out_chans,
                out_chans,
                kernel_size=3,
                padding=1,
                bias_attr=False,
            ),
            LayerNorm2d(out_chans),
        )

        self.net_2 = nn.Conv2D(
            256, 512, kernel_size=3, stride=2, padding=1, bias_attr=False
        )
        self.net_3 = nn.Conv2D(
            512, 1024, kernel_size=3, stride=2, padding=1, bias_attr=False
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.patch_embed(x)
        if self.pos_embed is not None:
            x = x + self.pos_embed
        for blk in self.blocks:
            x = blk(x)
        x = self.neck(x.transpose([0, 3, 1, 2]))
        x = self.net_2(x)
        x = self.net_3(x)
        return x


class Block(paddle.nn.Layer):
    """Transformer blocks with support of window attention and residual propagation blocks"""

    def __init__(
        self,
        dim: int,
        num_heads: int,
        mlp_ratio: float = 4.0,
        qkv_bias: bool = True,
        norm_layer: Type[nn.Layer] = nn.LayerNorm,
        act_layer: Type[nn.Layer] = nn.GELU,
        use_rel_pos: bool = False,
        rel_pos_zero_init: bool = True,
        window_size: int = 0,
        input_size: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Args:
            dim (int): Number of input channels.
            num_heads (int): Number of attention heads in each ViT block.
            mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
            qkv_bias (bool): If True, add a learnable bias to query, key, value.
            norm_layer (nn.Layer): Normalization layer.
            act_layer (nn.Layer): Activation layer.
            use_rel_pos (bool): If True, add relative positional embeddings to the attention map.
            rel_pos_zero_init (bool): If True, zero initialize relative positional parameters.
            window_size (int): Window size for window attention blocks. If it equals 0, then
                use global attention.
            input_size (tuple(int, int) or None): Input resolution for calculating the relative
                positional parameter size.
        """
        super().__init__()
        self.norm1 = norm_layer(dim)
        self.attn = Attention(
            dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            use_rel_pos=use_rel_pos,
            rel_pos_zero_init=rel_pos_zero_init,
            input_size=input_size if window_size == 0 else (window_size, window_size),
        )

        self.norm2 = norm_layer(dim)
        self.mlp = MLPBlock(
            embedding_dim=dim, mlp_dim=int(dim * mlp_ratio), act=act_layer
        )

        self.window_size = window_size

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        shortcut = x
        x = self.norm1(x)
        # Window partition
        if self.window_size > 0:
            H, W = x.shape[1], x.shape[2]
            x, pad_hw = window_partition(x, self.window_size)

        x = self.attn(x)
        # Reverse window partition
        if self.window_size > 0:
            x = window_unpartition(x, self.window_size, pad_hw, (H, W))

        x = shortcut + x
        x = x + self.mlp(self.norm2(x))

        return x


class Attention(paddle.nn.Layer):
    """Multi-head Attention block with relative position embeddings."""

    def __init__(
        self,
        dim: int,
        num_heads: int = 8,
        qkv_bias: bool = True,
        use_rel_pos: bool = False,
        rel_pos_zero_init: bool = True,
        input_size: Optional[Tuple[int, int]] = None,
    ) -> None:
        """
        Args:
            dim (int): Number of input channels.
            num_heads (int): Number of attention heads.
            qkv_bias (bool):  If True, add a learnable bias to query, key, value.
            rel_pos (bool): If True, add relative positional embeddings to the attention map.
            rel_pos_zero_init (bool): If True, zero initialize relative positional parameters.
            input_size (tuple(int, int) or None): Input resolution for calculating the relative
                positional parameter size.
        """
        super().__init__()
        self.num_heads = num_heads
        head_dim = dim // num_heads
        self.scale = head_dim**-0.5

        self.qkv = nn.Linear(dim, dim * 3, bias_attr=qkv_bias)
        self.proj = nn.Linear(dim, dim)

        self.use_rel_pos = use_rel_pos
        if self.use_rel_pos:
            assert (
                input_size is not None
            ), "Input size must be provided if using relative positional encoding."
            self.rel_pos_h = paddle.base.framework.EagerParamBase.from_tensor(
                tensor=paddle.zeros(shape=[2 * input_size[0] - 1, head_dim])
            )
            self.rel_pos_w = paddle.base.framework.EagerParamBase.from_tensor(
                tensor=paddle.zeros(shape=[2 * input_size[1] - 1, head_dim])
            )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        B, H, W, _ = tuple(x.shape)
        qkv = (
            self.qkv(x)
            .reshape([B, H * W, 3, self.num_heads, -1])
            .transpose([2, 0, 3, 1, 4])
        )
        q, k, v = qkv.reshape([3, B * self.num_heads, H * W, -1]).unbind(axis=0)

        attn = (q * self.scale) @ k.transpose([0, 2, 1])

        if self.use_rel_pos:
            attn = add_decomposed_rel_pos(
                attn, q, self.rel_pos_h, self.rel_pos_w, (H, W), (H, W)
            )

        attn = F.softmax(attn, axis=-1)
        x = (
            (attn @ v)
            .reshape([B, self.num_heads, H, W, -1])
            .transpose([0, 2, 3, 1, 4])
            .reshape([B, H, W, -1])
        )
        x = self.proj(x)

        return x


def window_partition(
    x: paddle.Tensor, window_size: int
) -> Tuple[paddle.Tensor, Tuple[int, int]]:
    """
    Partition into non-overlapping windows with padding if needed.
    Args:
        x (tensor): input tokens with [B, H, W, C].
        window_size (int): window size.

    Returns:
        windows: windows after partition with [B * num_windows, window_size, window_size, C].
        (Hp, Wp): padded height and width before partition
    """
    B, H, W, C = tuple(x.shape)

    pad_h = (window_size - H % window_size) % window_size
    pad_w = (window_size - W % window_size) % window_size
    if pad_h > 0 or pad_w > 0:
        x = F.pad(x, pad=(0, pad_w, 0, pad_h), data_format="NHWC")
    Hp, Wp = H + pad_h, W + pad_w

    x = x.reshape(
        [B, Hp // window_size, window_size, Wp // window_size, window_size, C]
    )
    windows = x.transpose([0, 1, 3, 2, 4, 5]).reshape([-1, window_size, window_size, C])
    return windows, (Hp, Wp)


def window_unpartition(
    windows: paddle.Tensor,
    window_size: int,
    pad_hw: Tuple[int, int],
    hw: Tuple[int, int],
) -> paddle.Tensor:
    """
    Window unpartition into original sequences and removing padding.
    Args:
        windows (tensor): input tokens with [B * num_windows, window_size, window_size, C].
        window_size (int): window size.
        pad_hw (Tuple): padded height and width (Hp, Wp).
        hw (Tuple): original height and width (H, W) before padding.

    Returns:
        x: unpartitioned sequences with [B, H, W, C].
    """
    Hp, Wp = pad_hw
    H, W = hw
    B = tuple(windows.shape)[0] // (Hp * Wp // window_size // window_size)
    x = windows.reshape(
        [B, Hp // window_size, Wp // window_size, window_size, window_size, -1]
    )
    x = x.transpose([0, 1, 3, 2, 4, 5]).reshape([B, Hp, Wp, -1])
    if Hp > H or Wp > W:
        x = x[:, :H, :W, :]
    return x


def get_rel_pos(q_size: int, k_size: int, rel_pos: paddle.Tensor) -> paddle.Tensor:
    """
    Get relative positional embeddings according to the relative positions of
        query and key sizes.
    Args:
        q_size (int): size of query q.
        k_size (int): size of key k.
        rel_pos (Tensor): relative position embeddings (L, C).

    Returns:
        Extracted positional embeddings according to relative positions.
    """
    max_rel_dist = int(2 * max(q_size, k_size) - 1)
    if tuple(rel_pos.shape)[0] != max_rel_dist:
        rel_pos_resized = paddle.nn.functional.interpolate(
            rel_pos.reshape([1, tuple(rel_pos.shape)[0], -1]).transpose([0, 2, 1]),
            size=max_rel_dist,
            mode="linear",
        )
        rel_pos_resized = rel_pos_resized.reshape([-1, max_rel_dist]).transpose([1, 0])
    else:
        rel_pos_resized = rel_pos

    q_coords = paddle.arange(end=q_size)[:, None] * max(k_size / q_size, 1.0)
    k_coords = paddle.arange(end=k_size)[None, :] * max(q_size / k_size, 1.0)
    relative_coords = q_coords - k_coords + (k_size - 1) * max(q_size / k_size, 1.0)
    return rel_pos_resized[relative_coords.astype(dtype="int64")]


def add_decomposed_rel_pos(
    attn: paddle.Tensor,
    q: paddle.Tensor,
    rel_pos_h: paddle.Tensor,
    rel_pos_w: paddle.Tensor,
    q_size: Tuple[int, int],
    k_size: Tuple[int, int],
) -> paddle.Tensor:
    """
    Calculate decomposed Relative Positional Embeddings from :paper:`mvitv2`.
    https://github.com/facebookresearch/mvit/blob/19786631e330df9f3622e5402b4a419a263a2c80/mvit/models/attention.py   # noqa B950
    Args:
        attn (Tensor): attention map.
        q (Tensor): query q in the attention layer with shape (B, q_h * q_w, C).
        rel_pos_h (Tensor): relative position embeddings (Lh, C) for height axis.
        rel_pos_w (Tensor): relative position embeddings (Lw, C) for width axis.
        q_size (Tuple): spatial sequence size of query q with (q_h, q_w).
        k_size (Tuple): spatial sequence size of key k with (k_h, k_w).

    Returns:
        attn (Tensor): attention map with added relative positional embeddings.
    """
    q_h, q_w = q_size
    k_h, k_w = k_size
    Rh = get_rel_pos(q_h, k_h, rel_pos_h)
    Rw = get_rel_pos(q_w, k_w, rel_pos_w)

    B, _, dim = tuple(q.shape)
    r_q = q.reshape([B, q_h, q_w, dim])
    rel_h = paddle.einsum("bhwc,hkc->bhwk", r_q, Rh)
    rel_w = paddle.einsum("bhwc,wkc->bhwk", r_q, Rw)

    attn = (
        attn.reshape([B, q_h, q_w, k_h, k_w])
        + rel_h[:, :, :, :, None]
        + rel_w[:, :, :, None, :]
    ).reshape([B, q_h * q_w, k_h * k_w])

    return attn


class PatchEmbed(paddle.nn.Layer):
    """
    Image to Patch Embedding.
    """

    def __init__(
        self,
        kernel_size: Tuple[int, int] = (16, 16),
        stride: Tuple[int, int] = (16, 16),
        padding: Tuple[int, int] = (0, 0),
        in_chans: int = 3,
        embed_dim: int = 768,
    ) -> None:
        """
        Args:
            kernel_size (Tuple): kernel size of the projection layer.
            stride (Tuple): stride of the projection layer.
            padding (Tuple): padding size of the projection layer.
            in_chans (int): Number of input image channels.
            embed_dim (int): Patch embedding dimension.
        """
        super().__init__()
        self.proj = nn.Conv2D(
            in_chans, embed_dim, kernel_size=kernel_size, stride=stride, padding=padding
        )

    def forward(self, x: paddle.Tensor) -> paddle.Tensor:
        x = self.proj(x)
        # B C H W -> B H W C
        x = x.transpose([0, 2, 3, 1])
        return x


DEFAULT_IMAGE_TOKEN = "<image>"
DEFAULT_IMAGE_PATCH_TOKEN = "<imgpad>"
DEFAULT_IM_START_TOKEN = "<img>"
DEFAULT_IM_END_TOKEN = "</img>"


class Qwen2LMHead(nn.Layer):
    def __init__(
        self,
        config,
        embedding_weights=None,
        transpose_y=False,
        tensor_parallel_output=1,
    ):
        super(Qwen2LMHead, self).__init__()
        self.config = config
        vocab_size = config.vocab_size

        self.transpose_y = transpose_y
        if transpose_y:
            # only for weight from embedding_weights
            if embedding_weights is not None:
                self.weight = embedding_weights
            else:
                self.weight = self.create_parameter(
                    shape=[vocab_size, config.hidden_size],
                    dtype=paddle.get_default_dtype(),
                )
        else:
            # for weight from model init
            self.weight = self.create_parameter(
                shape=[config.hidden_size, vocab_size],
                dtype=paddle.get_default_dtype(),
            )

    def forward(self, hidden_states, tensor_parallel_output=1):
        logits = paddle.matmul(hidden_states, self.weight, transpose_y=self.transpose_y)
        return logits


class GOTConfig(Qwen2Config):
    model_type = "GOT"


class GOTQwenModel(Qwen2Model):
    config_class = GOTConfig

    def __init__(self, config: Qwen2Config):
        super(GOTQwenModel, self).__init__(config)
        self.vision_tower_high = ImageEncoderViT(
            depth=12,
            embed_dim=768,
            img_size=1024,
            mlp_ratio=4,
            norm_layer=partial(paddle.nn.LayerNorm, epsilon=1e-6),
            num_heads=12,
            patch_size=16,
            qkv_bias=True,
            use_rel_pos=True,
            global_attn_indexes=[2, 5, 8, 11],
            window_size=14,
            out_chans=256,
        )
        self.mm_projector_vary = nn.Linear(1024, 1024)

    def forward(
        self,
        input_ids: paddle.Tensor = None,
        attention_mask: Optional[paddle.Tensor] = None,
        position_ids: Optional[paddle.Tensor] = None,
        past_key_values: Optional[List[paddle.Tensor]] = None,
        inputs_embeds: Optional[paddle.Tensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        images: Optional[paddle.Tensor] = None,
        return_dict: Optional[bool] = None,
    ):
        # HACK: replace back original embeddings for LLaVA pretraining
        orig_embeds_params = getattr(self, "orig_embeds_params", None)
        if orig_embeds_params is not None:
            with paddle.no_grad():
                self.get_input_embeddings().weight[: -self.num_new_tokens] = (
                    orig_embeds_params[: -self.num_new_tokens].data
                )

        if inputs_embeds is None:
            inputs_embeds = self.embed_tokens(input_ids)

        vision_tower_high = getattr(self, "vision_tower_high", None)

        if (
            vision_tower_high is not None
            and (input_ids.shape[1] != 1 or self.training)
            and images is not None
        ):
            use_im_start_end = getattr(self.config, "use_im_start_end", -1)

            im_patch_token = getattr(self.config, "im_patch_token", -1)
            im_start_token = getattr(self.config, "im_start_token", -1)
            im_end_token = getattr(self.config, "im_end_token", -1)

            im_patch_token = 151859
            im_start_token = 151857
            im_end_token = 151858

            image_features = []

            for image in images:
                if self.training:
                    image = image[1]
                P, C, H, W = image.shape
                if P == 1:
                    with paddle.set_grad_enabled(False):
                        cnn_feature = vision_tower_high(image)
                        cnn_feature = cnn_feature.flatten(2).transpose(
                            [0, 2, 1]
                        )  # 256*1024
                    image_feature = self.mm_projector_vary(cnn_feature)
                    image_features.append(image_feature)

                else:
                    image_patches = paddle.unbind(image)
                    image_patches_features = []
                    for image_patch in image_patches:
                        image_p = paddle.stack([image_patch])
                        with paddle.set_grad_enabled(False):
                            cnn_feature_p = vision_tower_high(image_p)
                            cnn_feature_p = cnn_feature_p.flatten(2).transpose(
                                [0, 2, 1]
                            )
                        image_feature_p = self.mm_projector_vary(cnn_feature_p)
                        image_patches_features.append(image_feature_p)
                    image_feature = paddle.concat(image_patches_features, axis=1)
                    image_features.append(image_feature)

            dummy_image_features_2 = paddle.zeros(
                [256, 1024], dtype=inputs_embeds.dtype
            )
            dummy_image_features = dummy_image_features_2
            use_im_start_end = True
            new_input_embeds = []
            for cur_input_ids, cur_input_embeds, cur_image_features in zip(
                input_ids, inputs_embeds, image_features
            ):
                if (cur_input_ids == im_patch_token).sum() == 0:
                    # multimodal LLM, but the current sample is not multimodal
                    cur_input_embeds = (
                        cur_input_embeds + (0.0 * dummy_image_features).sum()
                    )
                    new_input_embeds.append(cur_input_embeds)
                    continue

                if use_im_start_end:
                    if (cur_input_ids == im_start_token).sum() != (
                        cur_input_ids == im_end_token
                    ).sum():
                        raise ValueError(
                            "The number of image start tokens and image end tokens should be the same."
                        )

                    image_start_tokens = paddle.where(cur_input_ids == im_start_token)[
                        0
                    ]
                    for image_start_token_pos, per_cur_image_features in zip(
                        image_start_tokens, cur_image_features
                    ):
                        num_patches = per_cur_image_features.shape[0]

                        if (
                            cur_input_ids[image_start_token_pos + num_patches + 1]
                            != im_end_token
                        ):
                            raise ValueError(
                                "The image end token should follow the image start token."
                            )

                        cur_input_embeds = paddle.concat(
                            (
                                cur_input_embeds[: image_start_token_pos + 1],
                                per_cur_image_features,
                                cur_input_embeds[
                                    image_start_token_pos + num_patches + 1 :
                                ],
                            ),
                            axis=0,
                        )

                    new_input_embeds.append(cur_input_embeds)
                else:
                    raise NotImplementedError

            inputs_embeds = paddle.stack(new_input_embeds, axis=0)

        return super().forward(
            input_ids=None,
            attention_mask=attention_mask,
            past_key_values=past_key_values,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            position_ids=position_ids,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            return_dict=return_dict,
        )


class GOTQwenForCausalLM(Qwen2ForCausalLM):
    config_class = GOTConfig

    def __init__(self, config):
        super(Qwen2ForCausalLM, self).__init__(config)
        self.qwen2 = GOTQwenModel(config)

        self.vocab_size = config.vocab_size

        if config.tie_word_embeddings:
            self.lm_head = Qwen2LMHead(
                config,
                embedding_weights=self.qwen2.embed_tokens.weight,
                transpose_y=True,
            )
            self.tie_weights()
        else:
            self.lm_head = Qwen2LMHead(config)

        self.eval()

    def get_model(self):
        return self.qwen2

    def forward(
        self,
        input_ids: paddle.Tensor = None,
        attention_mask: Optional[paddle.Tensor] = None,
        position_ids: Optional[paddle.Tensor] = None,
        past_key_values: Optional[List[paddle.Tensor]] = None,
        inputs_embeds: Optional[paddle.Tensor] = None,
        labels: Optional[paddle.Tensor] = None,
        use_cache: Optional[bool] = None,
        output_attentions: Optional[bool] = None,
        output_hidden_states: Optional[bool] = None,
        images: Optional[paddle.Tensor] = None,
        return_dict: Optional[bool] = None,
    ):
        output_attentions = (
            output_attentions
            if output_attentions is not None
            else self.config.output_attentions
        )
        output_hidden_states = (
            output_hidden_states
            if output_hidden_states is not None
            else self.config.output_hidden_states
        )
        return_dict = (
            return_dict if return_dict is not None else self.config.use_return_dict
        )

        outputs = self.qwen2(
            input_ids=input_ids,
            past_key_values=past_key_values,
            attention_mask=attention_mask,
            position_ids=position_ids,
            inputs_embeds=inputs_embeds,
            use_cache=use_cache,
            output_attentions=output_attentions,
            output_hidden_states=output_hidden_states,
            images=images,
            return_dict=return_dict,
        )

        hidden_states = outputs[0]
        logits = self.lm_head(hidden_states)
        logits = logits.astype(dtype="float32")

        loss = None
        if labels is not None:
            # Shift so that tokens < n predict n
            shift_logits = logits[..., :-1, :]
            shift_labels = labels[..., 1:]
            loss_fct = nn.CrossEntropyLoss(reduction="sum")
            shift_logits = shift_logits.reshape([-1, self.config.vocab_size])
            shift_labels = shift_labels.reshape([-1])

            loss = loss_fct(shift_logits, shift_labels)
            label_sum = paddle.sum(shift_labels != -100)
            loss = loss / label_sum

        if not return_dict:
            output = (logits,) + outputs[1:]
            return (loss,) + output if loss is not None else output

        return CausalLMOutputWithPast(
            loss=loss,
            logits=logits,
            past_key_values=outputs.past_key_values,
            hidden_states=outputs.hidden_states,
            attentions=outputs.attentions,
        )

    def prepare_inputs_for_generation(
        self,
        input_ids,
        past_key_values=None,
        attention_mask=None,
        inputs_embeds=None,
        **kwargs
    ):
        batch_size, seq_length = input_ids.shape
        attention_mask = paddle.ones((batch_size, seq_length), dtype=paddle.bool)

        # Omit tokens covered by past_key_values
        if past_key_values is not None:
            past_length = past_key_values[0][0].shape[1]
            if past_length < input_ids.shape[1]:
                input_ids = input_ids[:, past_length:]

        position_ids = kwargs.get("position_ids", None)
        if attention_mask is not None and position_ids is None:
            # create position_ids on the fly for batch generation
            position_ids = attention_mask.astype(dtype="int64").cumsum(-1) - 1
            position_ids.masked_fill_(attention_mask == 0, 1)
            if past_key_values:
                position_ids = position_ids[:, -input_ids.shape[1] :]

        # if `inputs_embeds` are passed, we only want to use them in the 1st generation step
        if inputs_embeds is not None and past_key_values is None:
            model_inputs = {"inputs_embeds": inputs_embeds}
        else:
            model_inputs = {"input_ids": input_ids}

        model_inputs.update(
            {
                "position_ids": position_ids,
                "past_key_values": past_key_values,
                "use_cache": kwargs.get("use_cache"),
                "attention_mask": attention_mask,
                "images": kwargs.get("images", None),
            }
        )
        return model_inputs


class PPChart2TableInference(GOTQwenForCausalLM):

    def generate(self, inputs, **kwargs):
        max_new_tokens = kwargs.get("max_new_tokens", 1024)
        no_repeat_ngram_size = kwargs.get("no_repeat_ngram_size", 20)

        with paddle.no_grad():
            generated_ids = super().generate(
                inputs["input_ids"],
                images=inputs["images"],
                do_sample=False,
                num_beams=1,
                no_repeat_ngram_size=no_repeat_ngram_size,
                max_new_tokens=max_new_tokens,
            )

        return generated_ids



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
from pathlib import Path

from ...utils.misc import abspath
from ..image_classification import ClsTrainer
from .model_list import MODELS


class FaceRecTrainer(ClsTrainer):
    """Face Recognition Model Trainer"""

    entities = MODELS

    def update_config(self):
        """update training config"""
        if self.train_config.log_interval:
            self.pdx_config.update_log_interval(self.train_config.log_interval)
        if self.train_config.eval_interval:
            self.pdx_config.update_eval_interval(self.train_config.eval_interval)
        if self.train_config.save_interval:
            self.pdx_config.update_save_interval(self.train_config.save_interval)

        self.update_dataset_cfg()
        if self.train_config.num_classes is not None:
            self.pdx_config.update_num_classes(self.train_config.num_classes)
        if self.train_config.pretrain_weight_path != "":
            self.pdx_config.update_pretrained_weights(
                self.train_config.pretrain_weight_path
            )

        label_dict_path = Path(self.global_config.dataset_dir).joinpath("label.txt")
        if label_dict_path.exists():
            self.dump_label_dict(label_dict_path)
        if self.train_config.batch_size is not None:
            self.pdx_config.update_batch_size(self.train_config.batch_size)
        if self.train_config.learning_rate is not None:
            self.pdx_config.update_learning_rate(self.train_config.learning_rate)
        if self.train_config.epochs_iters is not None:
            self.pdx_config._update_epochs(self.train_config.epochs_iters)
        if self.train_config.warmup_steps is not None:
            self.pdx_config.update_warmup_epochs(self.train_config.warmup_steps)
        if self.global_config.output is not None:
            self.pdx_config._update_output_dir(self.global_config.output)

    def update_dataset_cfg(self):
        train_dataset_dir = abspath(
            os.path.join(self.global_config.dataset_dir, "train")
        )
        val_dataset_dir = abspath(os.path.join(self.global_config.dataset_dir, "val"))
        train_list_path = abspath(os.path.join(train_dataset_dir, "label.txt"))
        val_list_path = abspath(os.path.join(val_dataset_dir, "pair_label.txt"))

        ds_cfg = [
            f"DataLoader.Train.dataset.name=ClsDataset",
            f"DataLoader.Train.dataset.image_root={train_dataset_dir}",
            f"DataLoader.Train.dataset.cls_label_path={train_list_path}",
            f"DataLoader.Eval.dataset.name=FaceEvalDataset",
            f"DataLoader.Eval.dataset.dataset_root={val_dataset_dir}",
            f"DataLoader.Eval.dataset.pair_label_path={val_list_path}",
        ]
        self.pdx_config.update(ds_cfg)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Dict, List, Tuple, Union

import numpy as np

from ....modules.semantic_segmentation.model_list import MODELS
from ....utils.func_register import FuncRegister
from ...common.batch_sampler import ImageBatchSampler
from ...common.reader import ReadImage
from ..base import BasePredictor
from ..common import Normalize, ToBatch, ToCHWImage
from .processors import Resize, SegPostProcess
from .result import SegResult


class SegPredictor(BasePredictor):
    """SegPredictor that inherits from BasePredictor."""

    entities = MODELS

    _FUNC_MAP = {}
    register = FuncRegister(_FUNC_MAP)

    def __init__(
        self,
        target_size: Union[int, Tuple[int], None] = None,
        *args: List,
        **kwargs: Dict,
    ) -> None:
        """Initializes SegPredictor.

        Args:
            target_size: Image size used for inference.
            *args: Arbitrary positional arguments passed to the superclass.
            **kwargs: Arbitrary keyword arguments passed to the superclass.
        """
        super().__init__(*args, **kwargs)
        self.target_size = target_size
        self.preprocessors, self.infer, self.postprocessers = self._build()

    def _build_batch_sampler(self) -> ImageBatchSampler:
        """Builds and returns an ImageBatchSampler instance.

        Returns:
            ImageBatchSampler: An instance of ImageBatchSampler.
        """
        return ImageBatchSampler()

    def _get_result_class(self) -> type:
        """Returns the result class, SegResult.

        Returns:
            type: The SegResult class.
        """
        return SegResult

    def _build(self) -> Tuple:
        """Build the preprocessors, inference engine, and postprocessors based on the configuration.

        Returns:
            tuple: A tuple containing the preprocessors, inference engine, and postprocessors.
        """
        preprocessors = {"Read": ReadImage(format="RGB")}
        preprocessors["ToCHW"] = ToCHWImage()
        for cfg in self.config["Deploy"]["transforms"]:
            tf_key = cfg.pop("type")
            func = self._FUNC_MAP[tf_key]
            args = cfg
            name, op = func(self, **args) if args else func(self)
            preprocessors[name] = op
        preprocessors["ToBatch"] = ToBatch()
        if "Resize" not in preprocessors:
            _, op = self._FUNC_MAP["Resize"](self, target_size=-1)
            preprocessors["Resize"] = op

        if self.target_size is not None:
            _, op = self._FUNC_MAP["Resize"](self, target_size=self.target_size)
            preprocessors["Resize"] = op

        infer = self.create_static_infer()

        postprocessers = SegPostProcess()

        return preprocessors, infer, postprocessers

    def process(
        self,
        batch_data: List[Union[str, np.ndarray]],
        target_size: Union[int, Tuple[int], None] = None,
    ) -> Dict[str, Any]:
        """
        Process a batch of data through the preprocessing, inference, and postprocessing.

        Args:
            batch_data (List[Union[str, np.ndarray], ...]): A batch of input data (e.g., image file paths).
            target_size: Image size used for inference.

        Returns:
            dict: A dictionary containing the input path, raw image, and predicted segmentation maps for every instance of the batch. Keys include 'input_path', 'input_img', and 'pred'.
        """
        batch_raw_imgs = self.preprocessors["Read"](imgs=batch_data.instances)
        batch_imgs = self.preprocessors["Resize"](
            imgs=batch_raw_imgs, target_size=target_size
        )
        batch_imgs = self.preprocessors["Normalize"](imgs=batch_imgs)
        batch_imgs = self.preprocessors["ToCHW"](imgs=batch_imgs)
        x = self.preprocessors["ToBatch"](imgs=batch_imgs)
        batch_preds = self.infer(x=x)
        if len(batch_data) > 1:
            batch_preds = np.split(batch_preds[0], len(batch_data), axis=0)

        # postprocess
        batch_preds = self.postprocessers(batch_preds, batch_raw_imgs)

        return {
            "input_path": batch_data.input_paths,
            "page_index": batch_data.page_indexes,
            "input_img": batch_raw_imgs,
            "pred": batch_preds,
        }

    @register("Normalize")
    def build_normalize(
        self,
        mean=0.5,
        std=0.5,
    ):
        op = Normalize(mean=mean, std=std)
        return "Normalize", op

    @register("Resize")
    def build_resize(
        self,
        target_size=-1,
        keep_ratio=True,
        size_divisor=32,
        interp="LINEAR",
    ):
        op = Resize(
            target_size=target_size,
            keep_ratio=keep_ratio,
            size_divisor=size_divisor,
            interp=interp,
        )
        return "Resize", op



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
import importlib.util
import os
import unicodedata
from typing import Collection, Dict, List, Set, Tuple, Union

from .tokenizer_utils import PretrainedTokenizer
from .tokenizer_utils_base import AddedToken

__all__ = ["QWenTokenizer"]


VOCAB_FILES_NAMES = {"vocab_file": "qwen.tiktoken"}

PAT_STR = r"""(?i:'s|'t|'re|'ve|'m|'ll|'d)|[^\r\n\p{L}\p{N}]?\p{L}+|\p{N}| ?[^\s\p{L}\p{N}]+[\r\n]*|\s*[\r\n]+|\s+(?!\S)|\s+"""
ENDOFTEXT = "<|endoftext|>"
IMSTART = "<|im_start|>"
IMEND = "<|im_end|>"
# as the default behavior is changed to allow special tokens in
# regular texts, the surface forms of special tokens need to be
# as different as possible to minimize the impact
EXTRAS = tuple((f"<|extra_{i}|>" for i in range(205)))
SPECIAL_TOKENS = (
    ENDOFTEXT,
    IMSTART,
    IMEND,
) + EXTRAS

tiktoken = None


def is_tiktoken_available():
    return importlib.util.find_spec("tiktoken") is not None


def _load_tiktoken_bpe(tiktoken_bpe_file: str) -> Dict[bytes, int]:
    with open(tiktoken_bpe_file, "rb") as f:
        contents = f.read()
    return {
        base64.b64decode(token): int(rank)
        for token, rank in (line.split() for line in contents.splitlines() if line)
    }


class QWenTokenizer(PretrainedTokenizer):
    """QWen tokenizer."""

    model_input_names = ["input_ids", "attention_mask", "position_ids"]
    resource_files_names = VOCAB_FILES_NAMES

    def __init__(
        self,
        vocab_file,
        errors="replace",
        padding_side="left",
        **kwargs,
    ):
        super().__init__(**kwargs)
        if not is_tiktoken_available():
            raise ValueError(
                "tiktoken is not installed, please install it use: pip install tiktoken"
            )

        import tiktoken as tk

        tiktoken = tk

        self.errors = errors  # how to handle errors in decoding

        self.mergeable_ranks = _load_tiktoken_bpe(vocab_file)  # type: dict[bytes, int]
        self.special_tokens = {
            token: index
            for index, token in enumerate(
                SPECIAL_TOKENS, start=len(self.mergeable_ranks)
            )
        }

        enc = tiktoken.Encoding(
            "Qwen",
            pat_str=PAT_STR,
            mergeable_ranks=self.mergeable_ranks,
            special_tokens=self.special_tokens,
        )
        assert (
            len(self.mergeable_ranks) + len(self.special_tokens) == enc.n_vocab
        ), f"{len(self.mergeable_ranks) + len(self.special_tokens)} != {enc.n_vocab} in encoding"

        self.decoder = {
            v: k for k, v in self.mergeable_ranks.items()
        }  # type: dict[int, bytes|str]
        self.decoder.update({v: k for k, v in self.special_tokens.items()})

        self.tokenizer = enc  # type: tiktoken.Encoding

        self.eod_id = self.tokenizer.eot_token
        self.im_start_id = self.special_tokens[IMSTART]
        self.im_end_id = self.special_tokens[IMEND]

        if "pad_token_id" in kwargs:
            self.pad_token_id = kwargs["pad_token_id"]
        if "eos_token_id" in kwargs:
            self.eos_token_id = kwargs["eos_token_id"]

    def __len__(self) -> int:
        return self.tokenizer.n_vocab

    def get_vocab(self) -> Dict[bytes, int]:
        return self.mergeable_ranks

    def convert_tokens_to_ids(
        self, tokens: Union[bytes, str, List[Union[bytes, str]]]
    ) -> List[int]:
        ids = []
        if isinstance(tokens, (str, bytes)):
            if tokens in self.special_tokens:
                return self.special_tokens[tokens]
            else:
                return self.mergeable_ranks.get(tokens)
        for token in tokens:
            if token in self.special_tokens:
                ids.append(self.special_tokens[token])
            else:
                ids.append(self.mergeable_ranks.get(token))
        return ids

    def _update_tiktoken(self, tokens: List[str], special_tokens: bool = False) -> int:
        if special_tokens:
            added_tokens = []
            for token in tokens:
                if token in self.special_tokens:
                    continue

                token_id = len(self.mergeable_ranks) + len(self.special_tokens)
                self.special_tokens[token] = token_id
                self.decoder[token_id] = token

                added_tokens.append(token)

            import tiktoken

            self.tokenizer = tiktoken.Encoding(
                "Qwen",
                pat_str=PAT_STR,
                mergeable_ranks=self.mergeable_ranks,
                special_tokens=self.special_tokens,
            )

            return len(added_tokens)
        else:
            raise ValueError("Adding regular tokens is not supported")

    def _add_tokens(
        self,
        new_tokens: Union[List[str], List[AddedToken]],
        special_tokens: bool = False,
    ) -> int:
        if not special_tokens and new_tokens:
            raise ValueError("Adding regular tokens is not supported")
        new_tokens_str = []
        for token in new_tokens:
            surface_form = token.content if isinstance(token, AddedToken) else token
            new_tokens_str.append(surface_form)

        return self._update_tiktoken(new_tokens_str, special_tokens)

    def save_vocabulary(self, save_directory: str, **kwargs) -> Tuple[str]:
        """
        Save only the vocabulary of the tokenizer (vocabulary).

        Returns:
            `Tuple(str)`: Paths to the files saved.
        """
        file_path = os.path.join(save_directory, "qwen.tiktoken")
        with open(file_path, "w", encoding="utf8") as w:
            for k, v in self.mergeable_ranks.items():
                line = base64.b64encode(k).decode("utf8") + " " + str(v) + "\n"
                w.write(line)
        return (file_path,)

    def tokenize(
        self,
        text: str,
        allowed_special: Union[Set, str] = "all",
        disallowed_special: Union[Collection, str] = (),
        **kwargs,
    ) -> List[Union[bytes, str]]:
        """
        Converts a string in a sequence of tokens.

        Args:
            text (`str`):
                The sequence to be encoded.
            allowed_special (`Literal["all"]` or `set`):
                The surface forms of the tokens to be encoded as special tokens in regular texts.
                Default to "all".
            disallowed_special (`Literal["all"]` or `Collection`):
                The surface forms of the tokens that should not be in regular texts and trigger errors.
                Default to an empty tuple.

            kwargs (additional keyword arguments, *optional*):
                Will be passed to the underlying model specific encode method.

        Returns:
            `List[bytes|str]`: The list of tokens.
        """
        tokens = []
        text = unicodedata.normalize("NFC", text)

        # this implementation takes a detour: text -> token id -> token surface forms
        for t in self.tokenizer.encode(
            text, allowed_special=allowed_special, disallowed_special=disallowed_special
        ):
            tokens.append(self.decoder[t])
        return tokens

    def convert_tokens_to_string(self, tokens: List[Union[bytes, str]]) -> str:
        """
        Converts a sequence of tokens in a single string.
        """
        text = ""
        temp = b""
        for t in tokens:
            if isinstance(t, str):
                if temp:
                    text += temp.decode("utf-8", errors=self.errors)
                    temp = b""
                text += t
            elif isinstance(t, bytes):
                temp += t
            else:
                raise TypeError("token should only be of type types or str")
        if temp:
            text += temp.decode("utf-8", errors=self.errors)
        return text

    @property
    def vocab_size(self):
        return self.tokenizer.n_vocab

    def _convert_id_to_token(self, index: int) -> Union[bytes, str]:
        """Converts an id to a token, special tokens included"""
        if index in self.decoder:
            return self.decoder[index]
        raise ValueError("unknown ids")

    def _convert_token_to_id(self, token: Union[bytes, str]) -> int:
        """Converts a token to an id using the vocab, special tokens included"""
        if token in self.special_tokens:
            return self.special_tokens[token]
        if token in self.mergeable_ranks:
            return self.mergeable_ranks[token]
        raise ValueError("unknown token")

    def _tokenize(self, text: str, **kwargs):
        """
        Converts a string in a sequence of tokens (string), using the tokenizer. Split in words for word-based
        vocabulary or sub-words for sub-word-based vocabularies (BPE/SentencePieces/WordPieces).

        Do NOT take care of added tokens.
        """
        raise NotImplementedError

    def _decode(
        self,
        token_ids: Union[int, List[int]],
        skip_special_tokens: bool = False,
        errors: str = None,
        **kwargs,
    ) -> str:
        if isinstance(token_ids, int):
            token_ids = [token_ids]
        if skip_special_tokens:
            token_ids = [i for i in token_ids if i < self.eod_id]
        return self.tokenizer.decode(token_ids, errors=errors or self.errors)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from typing import List, Optional, Sequence, Tuple, Union

import numpy as np

from ....utils.deps import class_requires_deps, is_dep_available
from ...utils.benchmark import benchmark

if is_dep_available("opencv-contrib-python"):
    import cv2


@benchmark.timeit
@class_requires_deps("opencv-contrib-python")
class Scale:
    """Scale images."""

    def __init__(
        self,
        short_size: int,
        fixed_ratio: bool = True,
        keep_ratio: Union[bool, None] = None,
        do_round: bool = False,
    ) -> None:
        """
        Initializes the Scale class.

        Args:
            short_size (int): The target size for the shorter side of the image.
            fixed_ratio (bool): Whether to maintain a fixed aspect ratio of 4:3.
            keep_ratio (Union[bool, None]): Whether to keep the aspect ratio. Cannot be True if fixed_ratio is True.
            do_round (bool): Whether to round the scaling factor.
        """
        super().__init__()
        self.short_size = short_size
        assert (fixed_ratio and not keep_ratio) or (
            not fixed_ratio
        ), f"fixed_ratio and keep_ratio cannot be true at the same time"
        self.fixed_ratio = fixed_ratio
        self.keep_ratio = keep_ratio
        self.do_round = do_round

    def scale(self, video: List[np.ndarray]) -> List[np.ndarray]:
        """
        Performs resize operations on a sequence of images.

        Args:
            video (List[np.ndarray]): List where each item is an image,  as a numpy array.
             For example, [np.ndarray0, np.ndarray1, np.ndarray2, ...]

        Returns:
            List[np.ndarray]: List where each item is a np.ndarray after scaling.
        """

        imgs = video

        resized_imgs = []
        for i in range(len(imgs)):
            img = imgs[i]
            if isinstance(img, np.ndarray):
                h, w, _ = img.shape
            else:
                raise NotImplementedError
            if (w <= h and w == self.short_size) or (h <= w and h == self.short_size):
                resized_imgs.append(img)
                continue

            if w <= h:
                ow = self.short_size
                if self.fixed_ratio:
                    oh = int(self.short_size * 4.0 / 3.0)
                elif self.keep_ratio is False:
                    oh = self.short_size
                else:
                    scale_factor = self.short_size / w
                    oh = (
                        int(h * float(scale_factor) + 0.5)
                        if self.do_round
                        else int(h * self.short_size / w)
                    )
                    ow = (
                        int(w * float(scale_factor) + 0.5)
                        if self.do_round
                        else self.short_size
                    )
            else:
                oh = self.short_size
                if self.fixed_ratio:
                    ow = int(self.short_size * 4.0 / 3.0)
                elif self.keep_ratio is False:
                    ow = self.short_size
                else:
                    scale_factor = self.short_size / h
                    oh = (
                        int(h * float(scale_factor) + 0.5)
                        if self.do_round
                        else self.short_size
                    )
                    ow = (
                        int(w * float(scale_factor) + 0.5)
                        if self.do_round
                        else int(w * self.short_size / h)
                    )
            resized_imgs.append(
                cv2.resize(img, (ow, oh), interpolation=cv2.INTER_LINEAR)
            )
        imgs = resized_imgs
        return imgs

    def __call__(self, videos: List[np.ndarray]) -> List[np.ndarray]:
        """
        Apply the scaling operation to a list of videos.

        Args:
            videos (List[np.ndarray]): A list of videos, where each video is a sequence
            of images.

        Returns:
            List[np.ndarray]: A list of videos after scaling, where each video is a list of images.
        """
        return [self.scale(video) for video in videos]


@benchmark.timeit
class CenterCrop:
    """Center crop images."""

    def __init__(self, target_size: int, do_round: bool = True) -> None:
        """
        Initializes the CenterCrop class.

        Args:
            target_size (int): The size of the cropped area.
            do_round (bool): Whether to round the crop coordinates.
        """
        super().__init__()
        self.target_size = target_size
        self.do_round = do_round

    def center_crop(self, imgs: List[np.ndarray]) -> List[np.ndarray]:
        """
        Performs center crop operations on images.

        Args:
            imgs (List[np.ndarray]): A sequence of images (a numpy array).

        Returns:
            List[np.ndarray]: A list of images after center cropping or a cropped numpy array.
        """

        crop_imgs = []
        th, tw = self.target_size, self.target_size
        for img in imgs:
            h, w, _ = img.shape
            assert (w >= self.target_size) and (
                h >= self.target_size
            ), "image width({}) and height({}) should be larger than crop size".format(
                w, h, self.target_size
            )
            x1 = int(round((w - tw) / 2.0)) if self.do_round else (w - tw) // 2
            y1 = int(round((h - th) / 2.0)) if self.do_round else (h - th) // 2
            crop_imgs.append(img[y1 : y1 + th, x1 : x1 + tw])
        return crop_imgs

    def __call__(self, videos: List[np.ndarray]) -> List[np.ndarray]:
        """
        Apply the center crop operation to a list of videos.

        Args:
            videos (List[np.ndarray]): A list of videos, where each video is a sequence of images.

        Returns:
            List[np.ndarray]: A list of videos after center cropping.
        """
        return [self.center_crop(video) for video in videos]


@benchmark.timeit
class Image2Array:
    """Convert a sequence of images to a numpy array with optional transposition."""

    def __init__(self, transpose: bool = True, data_format: str = "tchw") -> None:
        """
        Initializes the Image2Array class.

        Args:
            transpose (bool): Whether to transpose the resulting numpy array.
            data_format (str): The format to transpose to, either 'tchw' or 'cthw'.

        Raises:
            AssertionError: If data_format is not one of the allowed values.
        """
        super().__init__()
        assert data_format in [
            "tchw",
            "cthw",
        ], f"Target format must in ['tchw', 'cthw'], but got {data_format}"
        self.transpose = transpose
        self.data_format = data_format

    def img2array(self, imgs: List[np.ndarray]) -> np.ndarray:
        """
        Converts a sequence of images to a numpy array and optionally transposes it.

        Args:
            imgs (List[np.ndarray]): A list of images to be converted to a numpy array.

        Returns:
            np.ndarray: A numpy array representation of the images.
        """
        t_imgs = np.stack(imgs).astype("float32")
        if self.transpose:
            if self.data_format == "tchw":
                t_imgs = t_imgs.transpose([0, 3, 1, 2])  # tchw
            else:
                t_imgs = t_imgs.transpose([3, 0, 1, 2])  # cthw
        return t_imgs

    def __call__(self, videos: List[np.ndarray]) -> List[np.ndarray]:
        """
        Apply the image to array conversion to a list of videos.

        Args:
            videos (List[Sequence[np.ndarray]]): A list of videos, where each video is a sequence of images.

        Returns:
            List[np.ndarray]: A list of numpy arrays, one for each video.
        """
        return [self.img2array(video) for video in videos]


@benchmark.timeit
@class_requires_deps("opencv-contrib-python")
class NormalizeVideo:
    """
    Normalize video frames by subtracting the mean and dividing by the standard deviation.
    """

    def __init__(
        self,
        mean: Sequence[float],
        std: Sequence[float],
        tensor_shape: Sequence[int] = [3, 1, 1],
        inplace: bool = False,
    ) -> None:
        """
        Initializes the NormalizeVideo class.

        Args:
            mean (Sequence[float]): The mean values for each channel.
            std (Sequence[float]): The standard deviation values for each channel.
            tensor_shape (Sequence[int]): The shape of the mean and std tensors.
            inplace (bool): Whether to perform normalization in place.
        """
        super().__init__()

        self.inplace = inplace
        if not inplace:
            self.mean = np.array(mean).reshape(tensor_shape).astype(np.float32)
            self.std = np.array(std).reshape(tensor_shape).astype(np.float32)
        else:
            self.mean = np.array(mean, dtype=np.float32)
            self.std = np.array(std, dtype=np.float32)

    def normalize_video(self, imgs: np.ndarray) -> np.ndarray:
        """
        Normalizes a sequence of images.

        Args:
            imgs (np.ndarray): A numpy array of images to be normalized.

        Returns:
            np.ndarray: The normalized images as a numpy array.
        """

        if self.inplace:
            n = len(imgs)
            h, w, c = imgs[0].shape
            norm_imgs = np.empty((n, h, w, c), dtype=np.float32)
            for i, img in enumerate(imgs):
                norm_imgs[i] = img

            for img in norm_imgs:  # [n,h,w,c]
                mean = np.float64(self.mean.reshape(1, -1))  # [1, 3]
                stdinv = 1 / np.float64(self.std.reshape(1, -1))  # [1, 3]
                cv2.subtract(img, mean, img)
                cv2.multiply(img, stdinv, img)
        else:
            imgs = imgs
            norm_imgs = imgs / 255.0
            norm_imgs -= self.mean
            norm_imgs /= self.std

        imgs = norm_imgs
        imgs = np.expand_dims(imgs, axis=0).copy()
        return imgs

    def __call__(self, videos: List[np.ndarray]) -> List[np.ndarray]:
        """
        Apply normalization to a list of videos.

        Args:
            videos (List[np.ndarray]): A list of videos, where each video is a numpy array of images.

        Returns:
            List[np.ndarray]: A list of normalized videos as numpy arrays.
        """
        return [self.normalize_video(video) for video in videos]


@benchmark.timeit
class VideoClasTopk:
    """Applies a top-k transformation on video classification predictions."""

    def __init__(self, class_ids: Optional[Sequence[Union[str, int]]] = None) -> None:
        """
        Initializes the VideoClasTopk class.

        Args:
            class_ids (Optional[Sequence[Union[str, int]]]): A list of class labels corresponding to class indices.
        """
        super().__init__()
        self.class_id_map = self._parse_class_id_map(class_ids)

    def softmax(self, data: np.ndarray) -> np.ndarray:
        """
        Applies the softmax function to an array of data.

        Args:
            data (np.ndarray): An array of data for which to compute softmax.

        Returns:
            np.ndarray: The softmax-transformed data.
        """
        x_max = np.max(data, axis=-1, keepdims=True)
        e_x = np.exp(data - x_max)
        return e_x / np.sum(e_x, axis=-1, keepdims=True)

    def _parse_class_id_map(
        self, class_ids: Optional[Sequence[Union[str, int]]]
    ) -> Optional[dict]:
        """
        Parses a list of class IDs into a mapping from class index to class label.

        Args:
            class_ids (Optional[Sequence[Union[str, int]]]): A list of class labels.

        Returns:
            Optional[dict]: A dictionary mapping class indices to labels, or None if no class_ids are provided.
        """
        if class_ids is None:
            return None
        class_id_map = {id: str(lb) for id, lb in enumerate(class_ids)}
        return class_id_map

    def __call__(
        self, preds: np.ndarray, topk: int = 5
    ) -> Tuple[np.ndarray, List[np.ndarray], List[List[str]]]:
        """
        Selects the top-k predictions from the classification output.

        Args:
            preds (np.ndarray): A 2D array of prediction scores.
            topk (int): The number of top predictions to return.

        Returns:
            Tuple[np.ndarray, List[np.ndarray], List[List[str]]]: A tuple containing:
                - An array of indices of the top-k predictions.
                - A list of arrays of scores for the top-k predictions.
                - A list of lists of label names for the top-k predictions.
        """
        preds[0] = self.softmax(preds[0])
        indexes = preds[0].argsort(axis=1)[:, -topk:][:, ::-1].astype("int32")
        scores = [
            list(np.around(pred[index], decimals=5))
            for pred, index in zip(preds[0], indexes)
        ]
        label_names = [[self.class_id_map[i] for i in index] for index in indexes]
        return indexes, scores, label_names


@benchmark.timeit
class ToBatch:
    """A class for batching videos."""

    def __call__(self, videos: List[np.ndarray]) -> List[np.ndarray]:
        """Call method to stack videos into a batch.

        Args:
            videos (list of np.ndarrays): List of videos to process.

        Returns:
            list of np.ndarrays: List containing a stacked tensor of the videos.
        """
        return [np.concatenate(videos, axis=0).astype(dtype=np.float32, copy=False)]



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import importlib.metadata
import os
import platform
import subprocess
import sys

from ..utils import logging
from ..utils.env import get_device_type

PLATFORM = platform.system()


def _check_call(*args, **kwargs):
    return subprocess.check_call(*args, **kwargs)


def _compare_version(version1, version2):
    import re

    def parse_version(version_str):
        version_pattern = re.compile(
            r"^(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)(?:-(?P<pre_release>.*))?(?:\+(?P<build_metadata>.+))?$"
        )
        match = version_pattern.match(version_str)
        if not match:
            raise ValueError(f"Unexpected version string: {version_str}")
        return (
            int(match.group("major")),
            int(match.group("minor")),
            int(match.group("patch")),
            match.group("pre_release"),
        )

    v1_infos = parse_version(version1)
    v2_infos = parse_version(version2)
    for v1_info, v2_info in zip(v1_infos, v2_infos):
        if v1_info is None and v2_info is None:
            continue
        if v1_info is None or (v2_info is not None and v1_info < v2_info):
            return -1
        if v2_info is None or (v1_info is not None and v1_info > v2_info):
            return 1
    return 0


def check_package_installation(package):
    try:
        importlib.metadata.distribution(package)
    except importlib.metadata.PackageNotFoundError:
        return False
    return True


def install_external_deps(repo_name, repo_root):
    """install paddle repository custom dependencies"""
    import paddle

    def get_gcc_version():
        return subprocess.check_output(["gcc", "--version"]).decode("utf-8").split()[2]

    if repo_name == "PaddleDetection":
        if os.path.exists(os.path.join(repo_root, "ppdet", "ext_op")):
            """Install custom op for rotated object detection"""
            if (
                PLATFORM == "Linux"
                and _compare_version(get_gcc_version(), "8.2.0") >= 0
                and "gpu" in get_device_type()
                and (
                    paddle.is_compiled_with_cuda()
                    and not paddle.is_compiled_with_rocm()
                )
            ):
                with switch_working_dir(os.path.join(repo_root, "ppdet", "ext_op")):
                    # TODO: Apply constraints here
                    args = [sys.executable, "setup.py", "install"]
                    _check_call(args)
            else:
                logging.warning(
                    "The custom operators in PaddleDetection for Rotated Object Detection is only supported when using CUDA, GCC>=8.2.0 and Paddle>=2.0.1, "
                    "your environment does not meet these requirements, so we will skip the installation of custom operators under PaddleDetection/ppdet/ext_ops, "
                    "which means you can not train the Rotated Object Detection models."
                )


def clone_repo_using_git(url, branch=None):
    """clone_repo_using_git"""
    args = ["git", "clone", "--depth", "1"]
    if isinstance(url, str):
        url = [url]
    args.extend(url)
    if branch is not None:
        args.extend(["-b", branch])
    return _check_call(args)


def fetch_repo_using_git(branch, url, depth=1):
    """fetch_repo_using_git"""
    args = ["git", "fetch", url, branch, "--depth", str(depth)]
    _check_call(args)


def reset_repo_using_git(pointer, hard=True):
    """reset_repo_using_git"""
    args = ["git", "reset", "--hard", pointer]
    return _check_call(args)


def remove_repo_using_rm(name):
    """remove_repo_using_rm"""
    if os.path.exists(name):
        if PLATFORM == "Windows":
            return _check_call(["rmdir", "/S", "/Q", name], shell=True)
        else:
            return _check_call(["rm", "-rf", name])


@contextlib.contextmanager
def mute():
    """mute"""
    with open(os.devnull, "w") as f:
        with contextlib.redirect_stdout(f), contextlib.redirect_stderr(f):
            yield


@contextlib.contextmanager
def switch_working_dir(new_wd):
    """switch_working_dir"""
    cwd = os.getcwd()
    os.chdir(new_wd)
    try:
        yield
    finally:
        os.chdir(cwd)



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
import PIL
from PIL import Image, ImageDraw, ImageFont

from ....utils.deps import class_requires_deps, is_dep_available
from ....utils.fonts import PINGFANG_FONT
from ...common.result import BaseVideoResult
from ...utils.color_map import get_colormap
from ...utils.io import VideoReader

if is_dep_available("opencv-contrib-python"):
    import cv2


@class_requires_deps("opencv-contrib-python")
class TopkVideoResult(BaseVideoResult):

    def _to_video(self):
        """Draw label on image"""
        labels = self.get("label_names", self["class_ids"])
        label_str = f"{labels[0]} {self['scores'][0]:.2f}"
        video_reader = VideoReader(backend="decord")
        video = video_reader.read(self["input_path"])
        video = list(video)
        write_fps = video_reader.get_fps()

        video_list = []
        for i in range(len(video)):
            image = Image.fromarray(video[i].asnumpy())
            image_size = image.size
            draw = ImageDraw.Draw(image)
            min_font_size = int(image_size[0] * 0.02)
            max_font_size = int(image_size[0] * 0.05)
            for font_size in range(max_font_size, min_font_size - 1, -1):
                font = ImageFont.truetype(
                    PINGFANG_FONT.path, font_size, encoding="utf-8"
                )
                if tuple(map(int, PIL.__version__.split("."))) <= (10, 0, 0):
                    text_width_tmp, text_height_tmp = draw.textsize(label_str, font)
                else:
                    left, top, right, bottom = draw.textbbox((0, 0), label_str, font)
                    text_width_tmp, text_height_tmp = right - left, bottom - top
                if text_width_tmp <= image_size[0]:
                    break
                else:
                    font = ImageFont.truetype(PINGFANG_FONT.path, min_font_size)
            color_list = get_colormap(rgb=True)
            color = tuple(color_list[0])
            font_color = tuple(self._get_font_colormap(3))
            if tuple(map(int, PIL.__version__.split("."))) <= (10, 0, 0):
                text_width, text_height = draw.textsize(label_str, font)
            else:
                left, top, right, bottom = draw.textbbox((0, 0), label_str, font)
                text_width, text_height = right - left, bottom - top

            rect_left = 3
            rect_top = 3
            rect_right = rect_left + text_width + 3
            rect_bottom = rect_top + text_height + 6

            draw.rectangle(
                [(rect_left, rect_top), (rect_right, rect_bottom)], fill=color
            )

            text_x = rect_left + 3
            text_y = rect_top
            draw.text((text_x, text_y), label_str, fill=font_color, font=font)
            image = cv2.cvtColor(np.array(image), cv2.COLOR_RGB2BGR)
            video_list.append(image)
        return {"res": (np.array(video_list), write_fps)}

    def _get_font_colormap(self, color_index):
        """
        Get font colormap
        """
        dark = np.array([0x14, 0x0E, 0x35])
        light = np.array([0xFF, 0xFF, 0xFF])
        light_indexs = [0, 3, 4, 8, 9, 13, 14, 18, 19]
        if color_index in light_indexs:
            return light.astype("int32")
        else:
            return dark.astype("int32")



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import os
from typing import Any, Dict, List, Tuple, Union

import pandas as pd

from ....modules.ts_classification.model_list import MODELS
from ...common.batch_sampler import TSBatchSampler
from ...common.reader import ReadTS
from ..base import BasePredictor
from ..common import BuildTSDataset, TSCutOff, TSNormalize, TStoArray, TStoBatch
from .processors import BuildPadMask, GetCls
from .result import TSClsResult


class TSClsPredictor(BasePredictor):
    """TSClsPredictor that inherits from BasePredictor."""

    entities = MODELS

    def __init__(self, *args: List, **kwargs: Dict) -> None:
        """Initializes TSClsPredictor.

        Args:
            *args: Arbitrary positional arguments passed to the superclass.
            **kwargs: Arbitrary keyword arguments passed to the superclass.
        """
        super().__init__(*args, **kwargs)
        self.preprocessors, self.infer, self.postprocessors = self._build()

    def _build_batch_sampler(self) -> TSBatchSampler:
        """Builds and returns an TSBatchSampler instance.

        Returns:
            TSBatchSampler: An instance of TSBatchSampler.
        """
        return TSBatchSampler()

    def _get_result_class(self) -> type:
        """Returns the result class.

        Returns:
            type: The Result class.
        """
        return TSClsResult

    def _build(self) -> Tuple:
        """Build the preprocessors, inference engine, and postprocessors based on the configuration.

        Returns:
            tuple: A tuple containing the preprocessors, inference engine, and postprocessors.
        """
        preprocessors = {
            "ReadTS": ReadTS(),
            "TSCutOff": TSCutOff(self.config["size"]),
        }

        if self.config.get("scale", None):
            scaler_file_path = os.path.join(self.model_dir, "scaler.pkl")
            if not os.path.exists(scaler_file_path):
                raise Exception(f"Cannot find scaler file: {scaler_file_path}")
            preprocessors["TSNormalize"] = TSNormalize(
                scaler_file_path, self.config["info_params"]
            )

        preprocessors["BuildTSDataset"] = BuildTSDataset(self.config["info_params"])
        preprocessors["BuildPadMask"] = BuildPadMask(self.config["input_data"])
        preprocessors["TStoArray"] = TStoArray(self.config["input_data"])
        preprocessors["TStoBatch"] = TStoBatch()
        infer = self.create_static_infer()
        postprocessors = {}
        postprocessors["GetCls"] = GetCls()
        return preprocessors, infer, postprocessors

    def process(self, batch_data: List[Union[str, pd.DataFrame]]) -> Dict[str, Any]:
        """
        Processes a batch of time series data through a series of preprocessing, inference, and postprocessing steps.

        Args:
            batch_data (List[Union[str, pd.DataFrame]]): A list of paths or identifiers for the batch of time series data to be processed.

        Returns:
            Dict[str, Any]: A dictionary containing the paths to the input data, the raw input time series, and the classification results.
        """
        batch_raw_ts = self.preprocessors["ReadTS"](ts_list=batch_data.instances)
        batch_raw_ts_ori = copy.deepcopy(batch_raw_ts)

        if "TSNormalize" in self.preprocessors:
            batch_ts = self.preprocessors["TSNormalize"](ts_list=batch_raw_ts)
            batch_input_ts = self.preprocessors["BuildTSDataset"](ts_list=batch_ts)
        else:
            batch_input_ts = self.preprocessors["BuildTSDataset"](ts_list=batch_raw_ts)

        batch_input_ts = self.preprocessors["BuildPadMask"](ts_list=batch_input_ts)
        batch_ts = self.preprocessors["TStoArray"](ts_list=batch_input_ts)

        x = self.preprocessors["TStoBatch"](ts_list=batch_ts)
        batch_preds = self.infer(x=x)

        batch_ts_preds = self.postprocessors["GetCls"](pred_list=batch_preds)

        return {
            "input_path": batch_data.input_paths,
            "input_ts": batch_raw_ts,
            "input_ts_data": batch_raw_ts_ori,
            "classification": batch_ts_preds,
            "target_cols": [self.config["info_params"]["target_cols"]],
        }



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Final, List, Optional, Tuple, Union

from pydantic import BaseModel

from ..infra.models import DataInfo, PrimaryOperations
from .shared import ocr

__all__ = [
    "INFER_ENDPOINT",
    "InferRequest",
    "LayoutParsingResult",
    "InferResult",
    "PRIMARY_OPERATIONS",
]

INFER_ENDPOINT: Final[str] = "/layout-parsing"


class InferRequest(ocr.BaseInferRequest):
    useDocOrientationClassify: Optional[bool] = None
    useDocUnwarping: Optional[bool] = None
    useTextlineOrientation: Optional[bool] = None
    useSealRecognition: Optional[bool] = None
    useTableRecognition: Optional[bool] = None
    useFormulaRecognition: Optional[bool] = None
    layoutThreshold: Optional[Union[float, dict]] = None
    layoutNms: Optional[bool] = None
    layoutUnclipRatio: Optional[Union[float, Tuple[float, float]]] = None
    layoutMergeBboxesMode: Optional[str] = None
    textDetLimitSideLen: Optional[int] = None
    textDetLimitType: Optional[str] = None
    textDetThresh: Optional[float] = None
    textDetBoxThresh: Optional[float] = None
    textDetUnclipRatio: Optional[float] = None
    textRecScoreThresh: Optional[float] = None
    sealDetLimitSideLen: Optional[int] = None
    sealDetLimitType: Optional[str] = None
    sealDetThresh: Optional[float] = None
    sealDetBoxThresh: Optional[float] = None
    sealDetUnclipRatio: Optional[float] = None
    sealRecScoreThresh: Optional[float] = None
    visualize: Optional[bool] = None


class LayoutParsingResult(BaseModel):
    prunedResult: dict
    outputImages: Optional[Dict[str, str]] = None
    inputImage: Optional[str] = None


class InferResult(BaseModel):
    layoutParsingResults: List[LayoutParsingResult]
    dataInfo: DataInfo


PRIMARY_OPERATIONS: Final[PrimaryOperations] = {
    "infer": (INFER_ENDPOINT, InferRequest, InferResult),
}



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
from typing import Dict

from ....utils.io import ImageReader, ImageWriter
from ..utils.mixin import ImgMixin, JsonMixin, StrMixin


class BaseResult(dict, StrMixin, JsonMixin):
    """Base Result"""

    def __init__(self, data: Dict) -> None:
        """Initializes the instance with the provided data.

        Args:
            data (Dict): The data to initialize the instance with.
        """
        super().__init__(data)
        self._show_funcs = []
        StrMixin.__init__(self)
        JsonMixin.__init__(self)

    def save_all(self, save_path: str) -> None:
        """
        Save all show functions to the specified path if they accept a save_path argument.

        Args:
            save_path (str): The path to save the functions' output.

        Returns:
            None
        """
        for func in self._show_funcs:
            signature = inspect.signature(func)
            if "save_path" in signature.parameters:
                func(save_path=save_path)
            else:
                func()


class CVResult(BaseResult, ImgMixin):
    """Result For Computer Vision Tasks"""

    def __init__(self, data: Dict) -> None:
        """Initializes the instance with the given data and sets up image processing with the 'pillow' backend.

        Args:
            data (Dict): The data to initialize the instance with.
        """
        super().__init__(data)
        ImgMixin.__init__(self, "pillow")
        self._img_reader = ImageReader(backend="pillow")
        self._img_writer = ImageWriter(backend="pillow")



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, List, Union

import numpy as np

from .setting import BLOCK_LABEL_MAP, LINE_SETTINGS
from .utils import (
    caculate_euclidean_dist,
    calculate_projection_overlap_ratio,
    is_english_letter,
    is_non_breaking_punctuation,
    is_numeric,
)

__all__ = [
    "TextSpan",
    "TextLine",
    "LayoutBlock",
    "LayoutRegion",
]


class TextSpan(object):
    """Text span class"""

    def __init__(self, box, text, label):
        """
        Initialize a TextSpan object.

        Args:
            box (list): The bounding box of the text span.
            text (str): The text content of the text span.
            label (int): The label of the text span.
        """
        self.box = box
        self.text = text
        self.label = label

    def __str__(self) -> str:
        return f"{self.text}"

    def __repr__(self) -> str:
        return f"{self.text}"


class TextLine(object):
    """Text line class"""

    def __init__(self, spans: List[TextSpan] = [], direction="horizontal"):
        """
        Initialize a TextLine object.

        Args:
            spans (List[TextSpan]): A list of TextSpan objects. Defaults to [].
            direction (str): The direction of the text line. Defaults to "horizontal".
        """
        self.spans = spans
        self.direction = direction
        self.region_box = self.get_region_box()
        self.need_new_line = False

    @property
    def labels(self):
        return [span.label for span in self.spans]

    @property
    def boxes(self):
        return [span.box for span in self.spans]

    @property
    def height(self):
        start_idx = 1 if self.direction == "horizontal" else 0
        end_idx = 3 if self.direction == "horizontal" else 2
        return abs(self.region_box[end_idx] - self.region_box[start_idx])

    @property
    def width(self):
        start_idx = 0 if self.direction == "horizontal" else 1
        end_idx = 2 if self.direction == "horizontal" else 3
        return abs(self.region_box[end_idx] - self.region_box[start_idx])

    def __str__(self) -> str:
        return f"{' '.join([str(span.text) for span in self.spans])}\n"

    def __repr__(self) -> str:
        return f"{' '.join([str(span.text) for span in self.spans])}\n"

    def add_span(self, span: Union[TextSpan, List[TextSpan]]):
        """
        Add a span to the text line.

        Args:
            span (Union[TextSpan, List[TextSpan]]): A single TextSpan object or a list of TextSpan objects.
        """
        if isinstance(span, list):
            self.spans.extend(span)
        else:
            self.spans.append(span)
        self.region_box = self.get_region_box()

    def get_region_box(self):
        """
        Get the region box of the text line.

        Returns:
            list: The region box of the text line.
        """
        if not self.spans:
            return None  # or an empty list, or however you want to handle no spans

        # Initialize min and max values with the first span's box
        x_min, y_min, x_max, y_max = self.spans[0].box

        for span in self.spans:
            x_min = min(x_min, span.box[0])
            y_min = min(y_min, span.box[1])
            x_max = max(x_max, span.box[2])
            y_max = max(y_max, span.box[3])

        return [x_min, y_min, x_max, y_max]

    def get_texts(
        self,
        block_label: str,
        block_text_width: int,
        block_start_coordinate: int,
        block_stop_coordinate: int,
        ori_image,
        text_rec_model=None,
        text_rec_score_thresh=None,
    ):
        """
        Get the text of the text line.

        Args:
            block_label (str): The label of the block.
            block_text_width (int): The width of the block.
            block_start_coordinate (int): The starting coordinate of the block.
            block_stop_coordinate (int): The stopping coordinate of the block.
            ori_image (np.ndarray): The original image.
            text_rec_model (Any): The text recognition model.
            text_rec_score_thresh (float): The text recognition score threshold.

        Returns:
            str: The text of the text line.
        """
        span_box_start_index = 0 if self.direction == "horizontal" else 1
        lines_start_index = 1 if self.direction == "horizontal" else 3
        self.spans.sort(
            key=lambda span: (
                span.box[span_box_start_index] // 2,
                (
                    span.box[lines_start_index]
                    if self.direction == "horizontal"
                    else -span.box[lines_start_index]
                ),
            )
        )
        if "formula" in self.labels:
            sort_index = 0 if self.direction == "horizontal" else 1
            splited_spans = self.split_boxes_by_projection()
            if len(self.spans) != len(splited_spans):
                splited_spans.sort(key=lambda span: span.box[sort_index])
                new_spans = []
                for span in splited_spans:
                    bbox = span.box
                    if span.label == "text":
                        crop_img = ori_image[
                            int(bbox[1]) : int(bbox[3]),
                            int(bbox[0]) : int(bbox[2]),
                        ]
                        crop_img_rec_res = list(text_rec_model([crop_img]))[0]
                        crop_img_rec_score = crop_img_rec_res["rec_score"]
                        crop_img_rec_text = crop_img_rec_res["rec_text"]
                        span.text = crop_img_rec_text
                        if crop_img_rec_score < text_rec_score_thresh:
                            continue
                    new_spans.append(span)
                self.spans = new_spans
        line_text = self.format_line(
            block_text_width,
            block_start_coordinate,
            block_stop_coordinate,
            line_gap_limit=self.height * 1.5,
            block_label=block_label,
        )
        return line_text

    def is_projection_contained(self, box_a, box_b, start_idx, end_idx):
        """Check if box_a completely contains box_b in the x-direction."""
        return box_a[start_idx] <= box_b[start_idx] and box_a[end_idx] >= box_b[end_idx]

    def split_boxes_by_projection(self, offset=1e-5):
        """
        Check if there is any complete containment in the x-direction
        between the bounding boxes and split the containing box accordingly.

        Args:
            offset (float): A small offset value to ensure that the split boxes are not too close to the original boxes.
        Returns:
            A new list of boxes, including split boxes, with the same `rec_text` and `label` attributes.
        """

        new_spans = []
        if self.direction == "horizontal":
            projection_start_index, projection_end_index = 0, 2
        else:
            projection_start_index, projection_end_index = 1, 3

        for i in range(len(self.spans)):
            span = self.spans[i]
            is_split = False
            for j in range(i, len(self.spans)):
                box_b = self.spans[j].box
                box_a, text, label = span.box, span.text, span.label
                if self.is_projection_contained(
                    box_a, box_b, projection_start_index, projection_end_index
                ):
                    is_split = True
                    # Split box_a based on the x-coordinates of box_b
                    if box_a[projection_start_index] < box_b[projection_start_index]:
                        w = (
                            box_b[projection_start_index]
                            - offset
                            - box_a[projection_start_index]
                        )
                        if w > 1:
                            new_bbox = box_a.copy()
                            new_bbox[projection_end_index] = (
                                box_b[projection_start_index] - offset
                            )
                            new_spans.append(
                                TextSpan(
                                    box=np.array(new_bbox),
                                    text=text,
                                    label=label,
                                )
                            )
                    if box_a[projection_end_index] > box_b[projection_end_index]:
                        w = (
                            box_a[projection_end_index]
                            - box_b[projection_end_index]
                            + offset
                        )
                        if w > 1:
                            box_a[projection_start_index] = (
                                box_b[projection_end_index] + offset
                            )
                            span = TextSpan(
                                box=np.array(box_a),
                                text=text,
                                label=label,
                            )
                if j == len(self.spans) - 1 and is_split:
                    new_spans.append(span)
            if not is_split:
                new_spans.append(span)

        return new_spans

    def format_line(
        self,
        block_text_width: int,
        block_start_coordinate: int,
        block_stop_coordinate: int,
        line_gap_limit: int = 10,
        block_label: str = "text",
    ) -> str:
        """
        Format a line of text spans based on layout constraints.

        Args:
            block_text_width (int): The width of the block.
            block_start_coordinate (int): The starting coordinate of the block.
            block_stop_coordinate (int): The stopping coordinate of the block.
            line_gap_limit (int): The limit for the number of pixels after the last span that should be considered part of the last line. Default is 10.
            block_label (str): The label associated with the entire block. Default is 'text'.
        Returns:
            str: Formatted line of text.
        """
        first_span_box = self.spans[0].box
        last_span_box = self.spans[-1].box

        line_text = ""
        for span in self.spans:
            if span.label == "formula" and block_label != "formula":
                formula_rec = span.text
                if not formula_rec.startswith("$") and not formula_rec.endswith("$"):
                    if len(self.spans) > 1:
                        span.text = f"${span.text}$"
                    else:
                        span.text = f"\n${span.text}$"
            line_text += span.text
            if (
                len(span.text) > 0
                and is_english_letter(line_text[-1])
                or span.label == "formula"
            ):
                line_text += " "

        if self.direction == "horizontal":
            text_stop_index = 2
        else:
            text_stop_index = 3

        if line_text.endswith(" "):
            line_text = line_text[:-1]

        if len(line_text) == 0:
            return ""

        last_char = line_text[-1]

        if (
            not is_english_letter(last_char)
            and not is_non_breaking_punctuation(last_char)
            and not is_numeric(last_char)
        ) or (
            block_stop_coordinate - last_span_box[text_stop_index]
            > block_text_width * 0.3
        ):
            if (
                self.direction == "horizontal"
                and block_stop_coordinate - last_span_box[text_stop_index]
                > line_gap_limit
            ) or (
                self.direction == "vertical"
                and (
                    block_stop_coordinate - last_span_box[text_stop_index]
                    > line_gap_limit
                    or first_span_box[1] - block_start_coordinate > line_gap_limit
                )
            ):
                self.need_new_line = True

        if line_text.endswith("-"):
            line_text = line_text[:-1]
            return line_text

        if (len(line_text) > 0 and is_english_letter(last_char)) or line_text.endswith(
            "$"
        ):
            line_text += " "
        if (
            len(line_text) > 0
            and not is_english_letter(last_char)
            and not is_numeric(last_char)
        ) or self.direction == "vertical":
            if (
                block_stop_coordinate - last_span_box[text_stop_index]
                > block_text_width * 0.3
                and len(line_text) > 0
                and not is_non_breaking_punctuation(last_char)
            ):
                line_text += "\n"
                self.need_new_line = True
        elif (
            block_stop_coordinate - last_span_box[text_stop_index]
            > (block_stop_coordinate - block_start_coordinate) * 0.5
        ):
            line_text += "\n"
            self.need_new_line = True

        return line_text


class LayoutBlock(object):
    """Layout Block Class"""

    def __init__(self, label, bbox, content="") -> None:
        """
        Initialize a LayoutBlock object.

        Args:
            label (str): Label assigned to the block.
            bbox (list): Bounding box coordinates of the block.
            content (str, optional): Content of the block. Defaults to an empty string.
        """
        self.label = label
        self.order_label = None
        self.bbox = list(map(int, bbox))
        self.content = content
        self.seg_start_coordinate = float("inf")
        self.seg_end_coordinate = float("-inf")
        self.width = bbox[2] - bbox[0]
        self.height = bbox[3] - bbox[1]
        self.area = float(self.width) * float(self.height)
        self.num_of_lines = 1
        self.image = None
        self.index = None
        self.order_index = None
        self.text_line_width = 1
        self.text_line_height = 1
        self.child_blocks = []
        self.update_direction()

    def __str__(self) -> str:
        _str = f"\n\n#################\nindex:\t{self.index}\nlabel:\t{self.label}\nregion_label:\t{self.order_label}\nbbox:\t{self.bbox}\ncontent:\t{self.content}\n#################"
        return _str

    def __repr__(self) -> str:
        _str = f"\n\n#################\nindex:\t{self.index}\nlabel:\t{self.label}\nregion_label:\t{self.order_label}\nbbox:\t{self.bbox}\ncontent:\t{self.content}\n#################"
        return _str

    def to_dict(self) -> dict:
        return self.__dict__

    def update_direction(self, direction=None) -> None:
        """
        Update the direction of the block based on its bounding box.

        Args:
            direction (str, optional): Direction of the block. If not provided, it will be determined automatically using the bounding box. Defaults to None.
        """
        if not direction:
            direction = self.get_bbox_direction()
        self.direction = direction
        self.update_direction_info()

    def update_direction_info(self) -> None:
        """Update the direction information of the block based on its direction."""
        if self.direction == "horizontal":
            self.secondary_direction = "vertical"
            self.short_side_length = self.height
            self.long_side_length = self.width
            self.start_coordinate = self.bbox[0]
            self.end_coordinate = self.bbox[2]
            self.secondary_direction_start_coordinate = self.bbox[1]
            self.secondary_direction_end_coordinate = self.bbox[3]
        else:
            self.secondary_direction = "horizontal"
            self.short_side_length = self.width
            self.long_side_length = self.height
            self.start_coordinate = self.bbox[1]
            self.end_coordinate = self.bbox[3]
            self.secondary_direction_start_coordinate = self.bbox[0]
            self.secondary_direction_end_coordinate = self.bbox[2]

    def append_child_block(self, child_block) -> None:
        """
        Append a child block to the current block.

        Args:
            child_block (LayoutBlock): Child block to be added.
        Returns:
            None
        """
        if not self.child_blocks:
            self.ori_bbox = self.bbox.copy()
        x1, y1, x2, y2 = self.bbox
        x1_child, y1_child, x2_child, y2_child = child_block.bbox
        union_bbox = (
            min(x1, x1_child),
            min(y1, y1_child),
            max(x2, x2_child),
            max(y2, y2_child),
        )
        self.bbox = union_bbox
        self.update_direction_info()
        child_blocks = [child_block]
        if child_block.child_blocks:
            child_blocks.extend(child_block.get_child_blocks())
        self.child_blocks.extend(child_blocks)

    def get_child_blocks(self) -> list:
        """Get all child blocks of the current block."""
        self.bbox = self.ori_bbox
        child_blocks = self.child_blocks.copy()
        self.child_blocks = []
        return child_blocks

    def get_centroid(self) -> tuple:
        """Get the centroid of the bounding box of the block."""
        x1, y1, x2, y2 = self.bbox
        centroid = ((x1 + x2) / 2, (y1 + y2) / 2)
        return centroid

    def get_bbox_direction(self, direction_ratio: float = 1.0) -> str:
        """
        Determine if a bounding box is horizontal or vertical.

        Args:
            direction_ratio (float): Ratio for determining direction. Default is 1.0.

        Returns:
            str: "horizontal" or "vertical".
        """
        return (
            "horizontal" if self.width * direction_ratio >= self.height else "vertical"
        )

    def calculate_text_line_direction(
        self, bboxes: List[List[int]], direction_ratio: float = 1.5
    ) -> bool:
        """
        Calculate the direction of the text based on the bounding boxes.

        Args:
            bboxes (list): A list of bounding boxes.
            direction_ratio (float): Ratio for determining direction. Default is 1.5.

        Returns:
            str: "horizontal" or "vertical".
        """

        horizontal_box_num = 0
        for bbox in bboxes:
            if len(bbox) != 4:
                raise ValueError(
                    "Invalid bounding box format. Expected a list of length 4."
                )
            x1, y1, x2, y2 = bbox
            width = x2 - x1
            height = y2 - y1
            horizontal_box_num += 1 if width * direction_ratio >= height else 0

        return "horizontal" if horizontal_box_num >= len(bboxes) * 0.5 else "vertical"

    def group_boxes_into_lines(
        self, ocr_rec_res, line_height_iou_threshold
    ) -> List[TextLine]:
        """
        Group the bounding boxes into lines based on their direction.

        Args:
            ocr_rec_res (dict): The result of OCR recognition.
            line_height_iou_threshold (float): The minimum IOU value required for two spans to belong to the same line.

        Returns:
            list: A list of TextLines.
        """
        rec_boxes = ocr_rec_res["boxes"]
        rec_texts = ocr_rec_res["rec_texts"]
        rec_labels = ocr_rec_res["rec_labels"]

        text_boxes = [
            rec_boxes[i] for i in range(len(rec_boxes)) if rec_labels[i] == "text"
        ]
        direction = self.calculate_text_line_direction(text_boxes)
        self.update_direction(direction)

        spans = [TextSpan(*span) for span in zip(rec_boxes, rec_texts, rec_labels)]

        if not spans:
            return []

        # sort spans by direction
        if self.direction == "vertical":
            spans.sort(
                key=lambda span: span.box[0], reverse=True
            )  # sort by x coordinate
            match_direction = "horizontal"
        else:
            spans.sort(
                key=lambda span: span.box[1], reverse=False
            )  # sort by y coordinate
            match_direction = "vertical"

        lines = []
        current_line = TextLine([spans[0]], direction=self.direction)

        for span in spans[1:]:
            overlap_ratio = calculate_projection_overlap_ratio(
                current_line.region_box, span.box, match_direction, mode="small"
            )

            if overlap_ratio >= line_height_iou_threshold:
                current_line.add_span(span)
            else:
                lines.append(current_line)
                current_line = TextLine([span], direction=self.direction)

        lines.append(current_line)

        if lines and self.direction == "vertical":
            line_heights = np.array([line.height for line in lines])
            min_height = np.min(line_heights)
            max_height = np.max(line_heights)

            # if height is too large, filter out the line
            if max_height > min_height * 2:
                normal_height_threshold = min_height * 1.1
                normal_height_count = np.sum(line_heights < normal_height_threshold)

                # if the number of lines with height less than the threshold is less than 40%, then filter out the line
                if normal_height_count < len(lines) * 0.4:
                    keep_condition = line_heights <= normal_height_threshold
                    lines = [line for line, keep in zip(lines, keep_condition) if keep]

        # calculate the average height of the text line
        if lines:
            line_heights = [line.height for line in lines]
            line_widths = [line.width for line in lines]
            self.text_line_height = np.mean(line_heights)
            self.text_line_width = np.mean(line_widths)
        else:
            self.text_line_height = 0
            self.text_line_width = 0

        return lines

    def update_text_content(
        self,
        image: list,
        ocr_rec_res: dict,
        text_rec_model: Any,
        text_rec_score_thresh: Union[float, None] = None,
    ) -> None:
        """
        Update the text content of the block based on the OCR result.

        Args:
            image (list): The input image.
            ocr_rec_res (dict): The result of OCR recognition.
            text_rec_model (Any): The model used for text recognition.
            text_rec_score_thresh (Union[float, None]): The score threshold for text recognition. If None, use the default setting.

        Returns:
            None
        """

        if len(ocr_rec_res["rec_texts"]) == 0:
            self.content = ""
            return

        lines = self.group_boxes_into_lines(
            ocr_rec_res,
            LINE_SETTINGS.get("line_height_iou_threshold", 0.8),
        )

        # words start coordinate and stop coordinate in the line
        coord_start_idx = 0 if self.direction == "horizontal" else 1
        coord_end_idx = coord_start_idx + 2

        if self.label == "reference":
            rec_boxes = ocr_rec_res["boxes"]
            block_start = min([box[coord_start_idx] for box in rec_boxes])
            block_stop = max([box[coord_end_idx] for box in rec_boxes])
        else:
            block_start = self.bbox[coord_start_idx]
            block_stop = self.bbox[coord_end_idx]

        text_lines = []
        text_width_list = []
        need_new_line_num = 0

        for line_idx, line in enumerate(lines):
            line: TextLine = line
            text_width_list.append(line.width)
            # get text from line
            line_text = line.get_texts(
                block_label=self.label,
                block_text_width=max(text_width_list),
                block_start_coordinate=block_start,
                block_stop_coordinate=block_stop,
                ori_image=image,
                text_rec_model=text_rec_model,
                text_rec_score_thresh=text_rec_score_thresh,
            )

            if line.need_new_line:
                need_new_line_num += 1

            # set segment start and end coordinate
            if line_idx == 0:
                self.seg_start_coordinate = line.spans[0].box[0]
            elif line_idx == len(lines) - 1:
                self.seg_end_coordinate = line.spans[-1].box[2]

            text_lines.append(line_text)

        delim = LINE_SETTINGS["delimiter_map"].get(self.label, "")

        if delim == "":
            content = ""
            pre_line_end = False
            last_char = ""
            for idx, line_text in enumerate(text_lines):
                if len(line_text) == 0:
                    continue

                line: TextLine = lines[idx]
                if pre_line_end:
                    start_gep_len = line.region_box[coord_start_idx] - block_start
                    if (
                        (
                            start_gep_len > line.height * 1.5
                            and not is_english_letter(last_char)
                            and not is_numeric(last_char)
                        )
                        or start_gep_len > (block_stop - block_start) * 0.4
                    ) and not content.endswith("\n"):
                        line_text = "\n" + line_text
                content += f"{line_text}"

                if len(line_text) > 2 and line_text.endswith(" "):
                    last_char = line_text[-2]
                else:
                    last_char = line_text[-1]
                if (
                    len(line_text) > 0
                    and not line_text.endswith("\n")
                    and not is_english_letter(last_char)
                    and not is_non_breaking_punctuation(last_char)
                    and not is_numeric(last_char)
                    and need_new_line_num > len(text_lines) * 0.5
                ) or need_new_line_num > len(text_lines) * 0.6:
                    content += f"\n"
                if (
                    block_stop - line.region_box[coord_end_idx]
                    > (block_stop - block_start) * 0.3
                ):
                    pre_line_end = True
        else:
            content = delim.join(text_lines)

        self.content = content
        self.num_of_lines = len(text_lines)


class LayoutRegion(LayoutBlock):
    """LayoutRegion class"""

    def __init__(
        self,
        bbox,
        blocks: List[LayoutBlock] = [],
    ) -> None:
        """
        Initialize a LayoutRegion object.

        Args:
            bbox (List[int]): The bounding box of the region.
            blocks (List[LayoutBlock]): A list of blocks that belong to this region.
        """
        super().__init__("region", bbox, content="")
        self.bbox = bbox
        self.block_map = {}
        self.direction = "horizontal"
        self.doc_title_block_idxes = []
        self.paragraph_title_block_idxes = []
        self.vision_block_idxes = []
        self.unordered_block_idxes = []
        self.vision_title_block_idxes = []
        self.normal_text_block_idxes = []
        self.euclidean_distance = float(np.inf)
        self.header_block_idxes = []
        self.footer_block_idxes = []
        self.text_line_width = 20
        self.text_line_height = 10
        self.num_of_lines = 10
        self.init_region_info_from_layout(blocks)
        self.update_euclidean_distance()

    def init_region_info_from_layout(self, blocks: List[LayoutBlock]) -> None:
        """Initialize the information about the layout region from the given blocks.

        Args:
            blocks (List[LayoutBlock]): A list of blocks that belong to this region.
        Returns:
            None
        """
        horizontal_normal_text_block_num = 0
        text_line_height_list = []
        text_line_width_list = []
        for idx, block in enumerate(blocks):
            self.block_map[idx] = block
            block.index = idx
            if block.label in BLOCK_LABEL_MAP["header_labels"]:
                self.header_block_idxes.append(idx)
            elif block.label in BLOCK_LABEL_MAP["doc_title_labels"]:
                self.doc_title_block_idxes.append(idx)
            elif block.label in BLOCK_LABEL_MAP["paragraph_title_labels"]:
                self.paragraph_title_block_idxes.append(idx)
            elif block.label in BLOCK_LABEL_MAP["vision_labels"]:
                self.vision_block_idxes.append(idx)
            elif block.label in BLOCK_LABEL_MAP["vision_title_labels"]:
                self.vision_title_block_idxes.append(idx)
            elif block.label in BLOCK_LABEL_MAP["footer_labels"]:
                self.footer_block_idxes.append(idx)
            elif block.label in BLOCK_LABEL_MAP["unordered_labels"]:
                self.unordered_block_idxes.append(idx)
            else:
                self.normal_text_block_idxes.append(idx)
                text_line_height_list.append(block.text_line_height)
                text_line_width_list.append(block.text_line_width)
                if block.direction == "horizontal":
                    horizontal_normal_text_block_num += 1
        direction = (
            "horizontal"
            if horizontal_normal_text_block_num
            >= len(self.normal_text_block_idxes) * 0.5
            else "vertical"
        )
        self.update_direction(direction)
        self.text_line_width = (
            np.mean(text_line_width_list) if text_line_width_list else 20
        )
        self.text_line_height = (
            np.mean(text_line_height_list) if text_line_height_list else 10
        )

    def update_euclidean_distance(self):
        """Update euclidean distance between each block and the reference point"""
        blocks: List[LayoutBlock] = list(self.block_map.values())
        if self.direction == "horizontal":
            ref_point = (0, 0)
            block_distance = [
                caculate_euclidean_dist((block.bbox[0], block.bbox[1]), ref_point)
                for block in blocks
            ]
        else:
            ref_point = (self.bbox[2], 0)
            block_distance = [
                caculate_euclidean_dist((block.bbox[2], block.bbox[1]), ref_point)
                for block in blocks
            ]
        self.euclidean_distance = min(block_distance) if len(block_distance) > 0 else 0

    def update_direction(self, direction=None):
        """
        Update the direction of the layout region.

        Args:
            direction (str): The new direction of the layout region.
        """
        super().update_direction(direction=direction)
        if self.direction == "horizontal":
            self.direction_start_index = 0
            self.direction_end_index = 2
            self.secondary_direction_start_index = 1
            self.secondary_direction_end_index = 3
            self.secondary_direction = "vertical"
        else:
            self.direction_start_index = 1
            self.direction_end_index = 3
            self.secondary_direction_start_index = 0
            self.secondary_direction_end_index = 2
            self.secondary_direction = "horizontal"

        self.direction_center_coordinate = (
            self.bbox[self.direction_start_index] + self.bbox[self.direction_end_index]
        ) / 2
        self.secondary_direction_center_coordinate = (
            self.bbox[self.secondary_direction_start_index]
            + self.bbox[self.secondary_direction_end_index]
        ) / 2



# Copyright (c) 2024 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import os.path as osp

from ...base.register import register_model_info, register_suite_info
from .config import TextDetConfig
from .model import TextDetModel
from .runner import TextDetRunner

REPO_ROOT_PATH = os.environ.get("PADDLE_PDX_PADDLEOCR_PATH")
PDX_CONFIG_DIR = osp.abspath(osp.join(osp.dirname(__file__), "..", "configs"))

register_suite_info(
    {
        "suite_name": "TextDet",
        "model": TextDetModel,
        "runner": TextDetRunner,
        "config": TextDetConfig,
        "runner_root_path": REPO_ROOT_PATH,
    }
)

################ Models Using Universal Config ################
register_model_info(
    {
        "model_name": "PP-OCRv4_mobile_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv4_mobile_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv4_server_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv4_server_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv4_server_seal_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv4_server_seal_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv4_mobile_seal_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv4_mobile_seal_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv3_mobile_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv3_mobile_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv3_server_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv3_server_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv5_server_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv5_server_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)

register_model_info(
    {
        "model_name": "PP-OCRv5_mobile_det",
        "suite": "TextDet",
        "config_path": osp.join(PDX_CONFIG_DIR, "PP-OCRv5_mobile_det.yaml"),
        "supported_apis": ["train", "evaluate", "predict", "export"],
    }
)
